from torch.optim.lr_scheduler import _LRScheduler


def polylr(optim, mdsteps, end_lr, power):
    '''
    Return PolynomialLRDecay scheduler object.

    Parameterts
    -----------
    optimizer : Optimizer
        optimizer object.

    mdsteps : int
        max steps for decay lr.

    end_lr : float
        last value for lr, no decay after that value.

    power : float
        decay value.
    '''
    return PolynomialLRDecay(optim, mdsteps, end_lr, power)


class PolynomialLRDecay(_LRScheduler):
    '''
    Class used to create custom PolynomialLRDecay scheduler.

    Attributes
    ----------
    max_decay_steps : int
        max steps for decay lr.

    end_learning_rate : float
        last value for lr, no decay after that value.

    power : float
        decay value.

    last_step : int
        current last step number.

    Methods
    -------
    get_lr()
        calculate current lr.

    step(step : int, optional)
        do scheduler step, change lr.
    '''

    def __init__(self, optimizer, max_decay_steps, end_learning_rate, power):
        '''
        Parameters
        ----------
        optimizer : Any
            optimizer object.

        max_decay_steps : int
            max steps for decay lr.

        end_learning_rate : float
            last value for lr, no decay after that value.

        power : float
            decay value.
        '''
        self.max_decay_steps = max_decay_steps
        self.end_learning_rate = end_learning_rate
        self.power = power
        self.last_step = 0
        super().__init__(optimizer)

    def get_lr(self):
        '''
        Calculate current lr.
        '''
        if self.last_step > self.max_decay_steps:
            return [self.end_learning_rate for _ in self.base_lrs]

        end_lr = self.end_learning_rate
        mult_val = ((1 - self.last_step / self.max_decay_steps)
                    ** (self.power))
        res_lst = [(base_lr - self.end_learning_rate) *
                   mult_val + end_lr for base_lr in self.base_lrs]

        return res_lst

    def step(self, step=None):
        '''
        Do scheduler step, change lr.

        Parameters
        ----------
        step : int, optional
            current step value.
        '''
        if step is None:
            step = self.last_step + 1
        self.last_step = step if step != 0 else 1

        if self.last_step <= self.max_decay_steps:
            end_lr = self.end_learning_rate
            mult_val = ((1 - self.last_step / self.max_decay_steps)
                        ** (self.power))
            decay_lrs = [(base_lr - self.end_learning_rate) * mult_val +
                         end_lr for base_lr in self.base_lrs]

            for param_group, lr in zip(self.optimizer.param_groups, decay_lrs):
                param_group['lr'] = lr



from .celoss import celoss, customceloss
from .celoss import CustomCrossEntropyLoss
from .bceloss import bce, bcelogits, custombce, custombcelogits
from .bceloss import CustomBCELoss, CustomBCEWithLogitsLoss
from .meanloss import mse, mae, custom_mse, custom_mae
from .meanloss import CustomMSELoss, CustomMAELoss
from .base import get_loss



from torch import nn


class ConvBlock(nn.Module):
    '''
    Conv block class.

    Methods
    -------
    forward(x)
        calculate features from input image x.
    '''

    def __init__(self, inc, outc, **kwargs):
        '''
        Create convolution block layers.

        Parameters
        ----------
        inc : int
            number of input channels.

        outc : int
            number of output channels.
        '''
        super().__init__()
        self.conv = nn.Conv2d(inc, outc, bias=False, **kwargs)
        self.bn = nn.BatchNorm2d(outc)
        self.relu = nn.ReLU()

    def forward(self, x):
        '''
        Calculate features from input image x.

        Parameterts
        -----------
        x : torch.tensor
            input image.
        '''
        return self.relu(self.bn(self.conv(x)))


class VGG11Features(nn.Module):
    '''
    Class used to calculate features using vgg-11 architecture.
    https://arxiv.org/pdf/1409.1556.pdf

    Methods
    -------
    forward(x)
        calculate features from input image x.

    out_features()
        return number of output features, needed for decoders input channels.

    name()
        return name of encoder.
    '''

    def __init__(self, inc):
        '''
        Create vgg-11 encoder layers.

        Parameters
        ----------
        inc : int
            number of input channels [1, 3, 4].
        '''
        super().__init__()
        self.features = 512
        self.conv1 = ConvBlock(inc, 64, kernel_size=(3, 3), padding=(1, 1))
        self.maxpool1 = nn.MaxPool2d(kernel_size=2, stride=2)
        self.conv2 = ConvBlock(64, 128, kernel_size=(3, 3), padding=(1, 1))
        self.maxpool2 = nn.MaxPool2d(kernel_size=2, stride=2)
        self.conv3 = ConvBlock(128, 256, kernel_size=(3, 3), padding=(1, 1))
        self.conv4 = ConvBlock(256, 256, kernel_size=(3, 3), padding=(1, 1))
        self.maxpool3 = nn.MaxPool2d(kernel_size=2, stride=2)
        self.conv5 = ConvBlock(256, 512, kernel_size=(3, 3), padding=(1, 1))
        self.conv6 = ConvBlock(512, 512, kernel_size=(3, 3), padding=(1, 1))
        self.maxpool4 = nn.MaxPool2d(kernel_size=2, stride=2)
        self.conv7 = ConvBlock(512, 512, kernel_size=(3, 3), padding=(1, 1))
        self.conv8 = ConvBlock(512, 512, kernel_size=(3, 3), padding=(1, 1))
        self.avgpool = nn.AdaptiveAvgPool2d(output_size=(1, 1))

    def forward(self, x):
        '''
        Calculate features from input image x.

        Parameterts
        -----------
        x : torch.tensor
            input image.
        '''
        b1 = self.maxpool1(self.conv1(x))
        b2 = self.maxpool2(self.conv2(b1))
        b3 = self.maxpool3(self.conv4(self.conv3(b2)))
        b4 = self.maxpool4(self.conv6(self.conv5(b3)))
        b5 = self.avgpool(self.conv8(self.conv7(b4)))
        return b5

    def out_features(self):
        '''
        Return number of output features, needed for decoders input channels.
        '''
        return self.features

    def name(self):
        '''
        Return name of encoder.
        '''
        return 'vgg11'


class VGG16Features(nn.Module):
    '''
    Class used to calculate features using vgg-16 architecture.
    https://arxiv.org/pdf/1409.1556.pdf

    Methods
    -------
    forward(x)
        calculate features from input image x.

    out_features()
        return number of output features, needed for decoders input channels.

    name()
        return name of encoder.
    '''

    def __init__(self, inc):
        '''
        Create vgg-16 encoder layers.

        Parameters
        ----------
        inc : int
            number of input channels [1, 3, 4].
        '''
        super().__init__()
        self.features = 512
        self.conv1 = ConvBlock(inc, 64, kernel_size=(3, 3), padding=(1, 1))
        self.conv2 = ConvBlock(64, 64, kernel_size=(3, 3), padding=(1, 1))
        self.maxpool1 = nn.MaxPool2d(kernel_size=2, stride=2)
        self.conv3 = ConvBlock(64, 128, kernel_size=(3, 3), padding=(1, 1))
        self.conv4 = ConvBlock(128, 128, kernel_size=(3, 3), padding=(1, 1))
        self.maxpool2 = nn.MaxPool2d(kernel_size=2, stride=2)
        self.conv5 = ConvBlock(128, 256, kernel_size=(3, 3), padding=(1, 1))
        self.conv6 = ConvBlock(256, 256, kernel_size=(3, 3), padding=(1, 1))
        self.conv7 = ConvBlock(256, 256, kernel_size=(1, 1))
        self.maxpool3 = nn.MaxPool2d(kernel_size=2, stride=2)
        self.conv8 = ConvBlock(256, 512, kernel_size=(3, 3), padding=(1, 1))
        self.conv9 = ConvBlock(512, 512, kernel_size=(3, 3), padding=(1, 1))
        self.conv10 = ConvBlock(512, 512, kernel_size=(1, 1))
        self.maxpool4 = nn.MaxPool2d(kernel_size=2, stride=2)
        self.conv11 = ConvBlock(512, 512, kernel_size=(3, 3), padding=(1, 1))
        self.conv12 = ConvBlock(512, 512, kernel_size=(3, 3), padding=(1, 1))
        self.conv13 = ConvBlock(512, 512, kernel_size=(1, 1), padding=(1, 1))
        self.avgpool = nn.AdaptiveAvgPool2d(output_size=(1, 1))

    def forward(self, x):
        '''
        Calculate features from input image x.

        Parameterts
        -----------
        x : torch.tensor
            input image.
        '''
        b1 = self.maxpool1(self.conv2(self.conv1(x)))
        b2 = self.maxpool2(self.conv4(self.conv3(b1)))
        b3 = self.maxpool3(self.conv7(self.conv6(self.conv5(b2))))
        b4 = self.maxpool4(self.conv10(self.conv9(self.conv8(b3))))
        b5 = self.avgpool(self.conv13(self.conv12(self.conv11(b4))))
        return b5

    def out_features(self):
        '''
        Return number of output features, needed for decoders input channels.
        '''
        return self.features

    def name(self):
        '''
        Return name of encoder.
        '''
        return 'vgg16'


class VGG19Features(nn.Module):
    '''
    Class used to calculate features using vgg-19 architecture.
    https://arxiv.org/pdf/1409.1556.pdf

    Methods
    -------
    forward(x)
        calculate features from input image x.

    out_features()
        return number of output features, needed for decoders input channels.

    name()
        return name of encoder.
    '''

    def __init__(self, inc):
        '''
        Create vgg-19 encoder layers.

        Parameters
        ----------
        inc : int
            number of input channels [1, 3, 4].
        '''
        super().__init__()
        self.features = 512
        self.conv1 = ConvBlock(inc, 64, kernel_size=(3, 3), padding=(1, 1))
        self.conv2 = ConvBlock(64, 64, kernel_size=(3, 3), padding=(1, 1))
        self.maxpool1 = nn.MaxPool2d(kernel_size=2, stride=2)
        self.conv3 = ConvBlock(64, 128, kernel_size=(3, 3), padding=(1, 1))
        self.conv4 = ConvBlock(128, 128, kernel_size=(3, 3), padding=(1, 1))
        self.maxpool2 = nn.MaxPool2d(kernel_size=2, stride=2)
        self.conv5 = ConvBlock(128, 256, kernel_size=(3, 3), padding=(1, 1))
        self.conv6 = ConvBlock(256, 256, kernel_size=(3, 3), padding=(1, 1))
        self.conv7 = ConvBlock(256, 256, kernel_size=(3, 3), padding=(1, 1))
        self.conv8 = ConvBlock(256, 256, kernel_size=(3, 3), padding=(1, 1))
        self.maxpool3 = nn.MaxPool2d(kernel_size=2, stride=2)
        self.conv9 = ConvBlock(256, 512, kernel_size=(3, 3), padding=(1, 1))
        self.conv10 = ConvBlock(512, 512, kernel_size=(3, 3), padding=(1, 1))
        self.conv11 = ConvBlock(512, 512, kernel_size=(3, 3), padding=(1, 1))
        self.conv12 = ConvBlock(512, 512, kernel_size=(3, 3), padding=(1, 1))
        self.maxpool4 = nn.MaxPool2d(kernel_size=2, stride=2)
        self.conv13 = ConvBlock(512, 512, kernel_size=(3, 3), padding=(1, 1))
        self.conv14 = ConvBlock(512, 512, kernel_size=(3, 3), padding=(1, 1))
        self.conv15 = ConvBlock(512, 512, kernel_size=(1, 1), padding=(1, 1))
        self.conv16 = ConvBlock(512, 512, kernel_size=(1, 1), padding=(1, 1))
        self.avgpool = nn.AdaptiveAvgPool2d(output_size=(1, 1))

    def forward(self, x):
        '''
        Calculate features from input image x.

        Parameterts
        -----------
        x : torch.tensor
            input image.
        '''
        b1 = self.maxpool1(self.conv2(self.conv1(x)))
        b2 = self.maxpool2(self.conv4(self.conv3(b1)))
        b3 = self.maxpool3(self.conv8(self.conv7(self.conv6(self.conv5(b2)))))
        b4 = self.maxpool4(self.conv12(self.conv11(
            self.conv10(self.conv9(b3)))))
        b5 = self.avgpool(self.conv16(self.conv15(
            self.conv14(self.conv13(b4)))))
        return b5

    def out_features(self):
        '''
        Return number of output features, needed for decoders input channels.
        '''
        return self.features

    def name(self):
        '''
        Return name of encoder.
        '''
        return 'vgg19'



from tqdm import tqdm
import torch
import torch.nn.functional as F
import numpy as np
import os
import matplotlib.pyplot as plt
import wandb


def train(train_dataloader,
          valid_dataloader,
          model,
          optimizer,
          scheduler,
          loss_fn,
          metrics,
          epochs,
          checkpoints,
          logs,
          device,
          save_every,
          one_hot,
          nc,
          wandb_flag,
          wandb_metrics):
    '''
    Train pipeline.

    Parameterts
    -----------
    train_dataloader : torch.utils.data.DataLoader
        train dataloader.

    valid_dataloader : torch.utils.data.DataLoader
        valid dataloader.

    model : Any
        pytorch model.

    optimizer : torch.optim.Optimizer
        optimizer for this model.

    scheduler : torch.optim.lr_scheduler
        scheduler for this optimizer.

    loss_fn : Any
        loss function.

    metrics : list[Any]
        list of metric functions.

    epochs : int
        epochs number.

    checkpoints : str
        folder for checkpoints.

    logs : str
        folder for logs.

    device : str
        device ['cpu', 'cuda'].

    save_every : int
        every save_every epoch save model weights.

    one_hot : bool
        one_hot for labels.

    nc : int
        classes number.

    wandb_flag : bool
        use wandb for logging.

    wandb_metrics: list[str]
        metric names which will be logged by wandb.
    '''
    checkpoints_folder = list(map(int, os.listdir(checkpoints)))
    checkpoints_folder = max(checkpoints_folder) + \
        1 if checkpoints_folder != [] else 0
    os.mkdir(f'{checkpoints}/{checkpoints_folder}', mode=0o777)
    checkpoints = f'{checkpoints}/{checkpoints_folder}'

    logs_folder = list(map(int, os.listdir(logs)))
    logs_folder = max(logs_folder) + 1 if logs_folder != [] else 0
    os.mkdir(f'{logs}/{logs_folder}', mode=0o777)
    logs = f'{logs}/{logs_folder}'

    tqdm_iter = tqdm(range(epochs))
    best_loss = 99999999.0
    best_epoch = 0

    range_daigrams = range(2 * (len(metrics) + 1) + 1)
    diagrams0 = [[] for _ in range_daigrams]
    diagrams1 = [[]for _ in range_daigrams]

    diagrams = [diagrams0, diagrams1, []]

    for ne, epoch in enumerate(tqdm_iter):
        model.train()
        train_loss = 0.0
        valid_loss = 0.0
        metric_names = [metric.name() for metric in metrics]
        train_mvct = [0.0] * len(metrics)
        valid_mvct = [0.0] * len(metrics)

        for batch in train_dataloader:
            imgs, labels = batch
            imgs = imgs.to(device)
            labels = labels.to(device)

            if one_hot:
                labels = F.one_hot(labels, nc).float()

            preds = model(imgs)
            loss = loss_fn(preds, labels)

            for mi, metric in enumerate(metrics):
                train_mvct[mi] += metric.calc(preds, labels).cpu()

            optimizer.zero_grad()
            loss.backward()
            optimizer.step()

            train_loss += loss.item()

        model.eval()
        with torch.no_grad():
            for batch in valid_dataloader:
                imgs, labels = batch
                imgs = imgs.to(device)
                labels = labels.to(device)

                if one_hot:
                    labels = F.one_hot(labels, nc).float()

                preds = model(imgs)
                loss = loss_fn(preds, labels)

                for mi, metric in enumerate(metrics):
                    valid_mvct[mi] += metric.calc(preds, labels).cpu()

                valid_loss += loss.item()

        train_mvct = np.array(train_mvct) / len(train_dataloader)
        valid_mvct = np.array(valid_mvct) / len(valid_dataloader)
        train_loss /= len(train_dataloader)
        valid_loss /= len(valid_dataloader)

        if wandb_flag:
            wandb.log({'train loss': train_loss,
                       'valid loss': valid_loss,
                       'learning rate': optimizer.param_groups[0]['lr']})

        if epoch % save_every == 0 or epoch == epochs - 1:
            state_dict = model.state_dict()
            torch.save(state_dict, checkpoints +
                       '/plan_{}_{:.2f}.pt'.format(epoch, valid_loss))

        if valid_loss < best_loss:
            best_epoch = epoch
            best_loss = valid_loss
            state_dict = model.state_dict()
            torch.save(state_dict, checkpoints + '/best.pt')

        tqdm_dct = {}
        train_str = ''
        valid_str = ''

        tqdm_dct['train loss'] = train_loss
        diagrams[1][0].append(train_loss)

        if ne == 0:
            diagrams[2].append('train_loss')

        for i, metric in enumerate(train_mvct):
            if wandb_metrics is not None and metric_names[i] in wandb_metrics:
                wandb.log({f'{metric_names[i]}': metric})

            diagrams[1][i + 1].append(metric)
            if ne == 0:
                diagrams[2].append(f'train_{metric_names[i]}')

            tqdm_dct[f'train_{metric_names[i]}'] = metric
            train_str += f'train_{metric_names[i]}: {metric} '

        tqdm_dct['valid loss'] = valid_loss
        diagrams[1][len(metrics) + 1].append(valid_loss)

        if ne == 0:
            diagrams[2].append('valid_loss')

        for i, metric in enumerate(valid_mvct):
            if wandb_metrics is not None and metric_names[i] in wandb_metrics:
                wandb.log({f'{metric_names[i]}': metric})

            diagrams[1][i + len(metrics) + 2].append(metric)
            if ne == 0:
                diagrams[2].append(f'valid_{metric_names[i]}')

            tqdm_dct[f'valid_{metric_names[i]}'] = metric
            if i != len(valid_mvct) - 1:
                valid_str += f'valid_{metric_names[i]}: {metric} '
            else:
                valid_str += f'valid_{metric_names[i]}: {metric}\n'

        tqdm_iter.set_postfix(tqdm_dct, refresh=True)

        with open(logs + '/trainval.log', 'a') as in_f:
            epoch_log = f'epoch: {epoch + 1}/{epochs} '
            loss_log = f'train loss: {train_loss} valid loss: {valid_loss} '
            metric_log = train_str + valid_str
            in_f.write(epoch_log + loss_log + metric_log)

        diagrams[1][-1].append(optimizer.param_groups[0]['lr'])
        for k in range(len(diagrams[0])):
            diagrams[0][k].append(ne)
        if ne == 0:
            diagrams[2].append('lr')

        scheduler.step()
        tqdm_iter.refresh()

    for ind, name in enumerate(diagrams[2]):
        plt.plot(diagrams[0][ind], diagrams[1][ind])
        plt.savefig(logs + f'/{name}.png')
        plt.close()

    os.rename(checkpoints + '/best.pt', checkpoints +
              '/best_{}_{:.2f}.pt'.format(best_epoch, best_loss))


def test(test_dataloader,
         model,
         loss_fn,
         metrics,
         logs,
         device,
         one_hot,
         nc,
         wandb_flag,
         wandb_metrics):
    '''
    Test pipeline.

    Parameterts
    -----------
    test_dataloader : torch.utils.data.DataLoader
        test dataloader.

    model : Any
        pytorch model.

    loss_fn : Any
        loss function.

    metrics : list[Any]
        list of metric functions.

    logs : str
        folder for logs.

    device : str
        device ['cpu', 'cuda'].

    one_hot : bool
        one_hot for labels.

    nc : int
        classes number.

    wandb_flag : bool
        use wandb for logging.

    wandb_metrics: list[str]
        metric names which will be logged by wandb.
    '''
    logs_folder = list(map(int, os.listdir(logs)))
    logs_folder = max(logs_folder) + 1 if logs_folder != [] else 0
    os.mkdir(f'{logs}/{logs_folder}', mode=0o777)
    logs = f'{logs}/{logs_folder}'

    model.eval()
    test_loss = 0.0
    metric_names = [metric.name() for metric in metrics]
    test_mvct = [0.0] * len(metrics)

    with torch.no_grad():
        for batch in test_dataloader:
            imgs, labels = batch
            imgs = imgs.to(device)
            labels = labels.to(device)

            if one_hot:
                labels = F.one_hot(labels, nc).float()

            preds = model(imgs)
            loss = loss_fn(preds, labels)

            for mi, metric in enumerate(metrics):
                test_mvct[mi] += metric.calc(preds, labels).cpu()

            test_loss += loss.item()

    test_mvct = np.array(test_mvct) / len(test_dataloader)
    test_loss /= len(test_dataloader)
    test_str = ''

    if wandb_flag:
        wandb.log({'test loss': test_loss})

    for i, metric in enumerate(test_mvct):
        if wandb_metrics is not None and metric_names[i] in wandb_metrics:
            wandb.log({f'{metric_names[i]}': metric})

        if i != len(test_mvct) - 1:
            test_str += f'{metric_names[i]}: {metric} '
        else:
            test_str += f'{metric_names[i]}: {metric}\n'

    with open(logs + '/test.log', 'a') as in_f:
        in_f.write(f'test loss: {test_loss} ' + test_str)

    return logs



from torch import nn


class ResidualBlock1(nn.Module):
    '''
    Residual block class type 1.

    Methods
    -------
    forward(x)
        calculate features from input image x.
    '''

    def __init__(self, inc, outc):
        '''
        Create residual block type 1 layers.

        Parameters
        ----------
        inc : int
            number of input channels.

        outc : int
            number of output channels.
        '''
        super().__init__()
        if outc == inc:
            self.conv1 = nn.Conv2d(inc,
                                   outc,
                                   kernel_size=(3, 3),
                                   stride=(1, 1),
                                   padding=(1, 1),
                                   bias=False)
            self.downsample = None
        else:
            self.conv1 = nn.Conv2d(inc,
                                   outc,
                                   kernel_size=(3, 3),
                                   stride=(2, 2),
                                   padding=(1, 1),
                                   bias=False)
            self.downsample = Downsample(inc, outc)
        self.bn1 = nn.BatchNorm2d(outc)
        self.conv2 = nn.Conv2d(outc,
                               outc,
                               kernel_size=(3, 3),
                               stride=(1, 1),
                               padding=(1, 1),
                               bias=False)
        self.bn2 = nn.BatchNorm2d(outc)
        self.relu = nn.ReLU()

    def forward(self, x):
        '''
        Calculate features from input image x.

        Parameterts
        -----------
        x : torch.tensor
            input image.
        '''
        b1 = self.relu(self.bn1(self.conv1(x)))
        b2 = self.bn2(self.conv2(b1))
        if self.downsample is not None:
            x = self.downsample(x)
        return x + b2


class ResidualBlock2(nn.Module):
    '''
    Residual block class type 2.

    Methods
    -------
    forward(x)
        calculate features from input image x.
    '''

    def __init__(self, inc, outc):
        '''
        Create residual block type 2 layers.

        Parameters
        ----------
        inc : int
            number of input channels.

        outc : int
            number of output channels.
        '''
        super().__init__()
        self.conv1 = nn.Conv2d(inc,
                               outc,
                               kernel_size=(1, 1),
                               stride=(1, 1),
                               bias=False)
        self.bn1 = nn.BatchNorm2d(outc)
        if inc == outc or inc == int(outc * 4):
            self.conv2 = nn.Conv2d(outc,
                                   outc,
                                   kernel_size=(3, 3),
                                   stride=(1, 1),
                                   padding=(1, 1),
                                   bias=False)
            tmp_down = Downsample(inc, int(outc * 4), stride=(1, 1))
            self.downsample = tmp_down if inc == outc else None
        else:
            self.conv2 = nn.Conv2d(outc,
                                   outc,
                                   kernel_size=(3, 3),
                                   stride=(2, 2),
                                   padding=(1, 1),
                                   bias=False)
            self.downsample = Downsample(inc, int(outc * 4))
        self.bn2 = nn.BatchNorm2d(outc)
        self.conv3 = nn.Conv2d(outc,
                               int(outc * 4),
                               kernel_size=(1, 1),
                               stride=(1, 1),
                               bias=False)
        self.bn3 = nn.BatchNorm2d(int(outc * 4))
        self.relu = nn.ReLU()

    def forward(self, x):
        '''
        Calculate features from input image x.

        Parameterts
        -----------
        x : torch.tensor
            input image.
        '''
        b1 = self.bn1(self.conv1(x))
        b2 = self.bn2(self.conv2(b1))
        b3 = self.relu(self.bn3(self.conv3(b2)))
        if self.downsample is not None:
            x = self.downsample(x)
        return x + b3


class Downsample(nn.Module):
    '''
    Downsample block class.

    Methods
    -------
    forward(x)
        calculate features from input image x.
    '''

    def __init__(self, inc, outc, stride=(2, 2)):
        '''
        Create downsample block layers.

        Parameters
        ----------
        inc : int
            number of input channels.

        outc : int
            number of output channels.

        stride: tuple(int, int), int, optional
            stride for convolution layer.
        '''
        super().__init__()
        self.conv = nn.Conv2d(inc,
                              outc,
                              kernel_size=(1, 1),
                              stride=stride,
                              bias=False)
        self.bn = nn.BatchNorm2d(outc)

    def forward(self, x):
        '''
        Calculate features from input image x.

        Parameterts
        -----------
        x : torch.tensor
            input image.
        '''
        return self.bn(self.conv(x))


class Resnet18Features(nn.Module):
    '''
    Class used to calculate features using resnet18 architecture.
    https://arxiv.org/pdf/1512.03385.pdf

    Methods
    -------
    forward(x)
        calculate features from input image x.

    out_features()
        return number of output features, needed for decoders input channels.

    name()
        return name of encoder.
    '''

    def __init__(self, inc):
        '''
        Create resnet18 encoder layers.

        Parameters
        ----------
        inc : int
            number of input channels [1, 3, 4].
        '''
        super().__init__()
        self.features = 512
        self.conv1 = nn.Conv2d(inc,
                               64,
                               kernel_size=(7, 7),
                               stride=(2, 2),
                               padding=(3, 3),
                               bias=False)
        self.bn1 = nn.BatchNorm2d(64)
        self.relu = nn.ReLU()
        self.maxpool = nn.MaxPool2d(kernel_size=3,
                                    stride=2,
                                    padding=1,
                                    dilation=1,
                                    ceil_mode=False)
        self.resblock11 = ResidualBlock1(64, 64)
        self.resblock12 = ResidualBlock1(64, 64)
        self.resblock21 = ResidualBlock1(64, 128)
        self.resblock22 = ResidualBlock1(128, 128)
        self.resblock31 = ResidualBlock1(128, 256)
        self.resblock32 = ResidualBlock1(256, 256)
        self.resblock41 = ResidualBlock1(256, 512)
        self.resblock42 = ResidualBlock1(512, 512)
        self.avgpool = nn.AdaptiveAvgPool2d(output_size=(1, 1))

    def forward(self, x):
        '''
        Calculate features from input image x.

        Parameterts
        -----------
        x : torch.tensor
            input image.
        '''
        pr = self.maxpool(self.relu(self.bn1(self.conv1(x))))
        b1 = self.resblock12(self.resblock11(pr))
        b2 = self.resblock22(self.resblock21(b1))
        b3 = self.resblock32(self.resblock31(b2))
        b4 = self.resblock42(self.resblock41(b3))
        b5 = self.avgpool(b4)
        return b5

    def out_features(self):
        '''
        Return number of output features, needed for decoders input channels.
        '''
        return self.features

    def name(self):
        '''
        Return name of encoder.
        '''
        return 'resnet18'


class Resnet34Features(nn.Module):
    '''
    Class used to calculate features using resnet34 architecture.
    https://arxiv.org/pdf/1512.03385.pdf

    Methods
    -------
    forward(x)
        calculate features from input image x.

    out_features()
        return number of output features, needed for decoders input channels.

    name()
        return name of encoder.
    '''

    def __init__(self, inc):
        '''
        Create resnet34 encoder layers.

        Parameters
        ----------
        inc : int
            number of input channels [1, 3, 4].
        '''
        super().__init__()
        self.features = 512
        self.conv1 = nn.Conv2d(inc,
                               64,
                               kernel_size=(7, 7),
                               stride=(2, 2),
                               padding=(3, 3),
                               bias=False)
        self.bn1 = nn.BatchNorm2d(64)
        self.relu = nn.ReLU()
        self.maxpool = nn.MaxPool2d(kernel_size=3,
                                    stride=2,
                                    padding=1,
                                    dilation=1,
                                    ceil_mode=False)
        self.resblock11 = ResidualBlock1(64, 64)
        self.resblock12 = ResidualBlock1(64, 64)
        self.resblock13 = ResidualBlock1(64, 64)
        self.resblock21 = ResidualBlock1(64, 128)
        self.resblock22 = ResidualBlock1(128, 128)
        self.resblock23 = ResidualBlock1(128, 128)
        self.resblock24 = ResidualBlock1(128, 128)
        self.resblock31 = ResidualBlock1(128, 256)
        self.resblock32 = ResidualBlock1(256, 256)
        self.resblock33 = ResidualBlock1(256, 256)
        self.resblock34 = ResidualBlock1(256, 256)
        self.resblock35 = ResidualBlock1(256, 256)
        self.resblock36 = ResidualBlock1(256, 256)
        self.resblock41 = ResidualBlock1(256, 512)
        self.resblock42 = ResidualBlock1(512, 512)
        self.resblock43 = ResidualBlock1(512, 512)
        self.avgpool = nn.AdaptiveAvgPool2d(output_size=(1, 1))

    def forward(self, x):
        '''
        Calculate features from input image x.

        Parameterts
        -----------
        x : torch.tensor
            input image.
        '''
        pr = self.maxpool(self.relu(self.bn1(self.conv1(x))))
        b1 = self.resblock13(self.resblock12(self.resblock11(pr)))
        b2 = self.resblock24(self.resblock23(
            self.resblock22(self.resblock21(b1))))
        b3 = self.resblock36(self.resblock35(self.resblock34(
            self.resblock33(self.resblock32(self.resblock31(b2))))))
        b4 = self.resblock43(self.resblock42(self.resblock41(b3)))
        b5 = self.avgpool(b4)
        return b5

    def out_features(self):
        '''
        Return number of output features, needed for decoders input channels.
        '''
        return self.features

    def name(self):
        '''
        Return name of encoder.
        '''
        return 'resnet34'


class Resnet50Features(nn.Module):
    '''
    Class used to calculate features using resnet50 architecture.
    https://arxiv.org/pdf/1512.03385.pdf

    Methods
    -------
    forward(x)
        calculate features from input image x.

    out_features()
        return number of output features, needed for decoders input channels.

    name()
        return name of encoder.
    '''

    def __init__(self, inc):
        '''
        Create resnet50 encoder layers.

        Parameters
        ----------
        inc : int
            number of input channels [1, 3, 4].
        '''
        super().__init__()
        self.features = 2048
        self.conv1 = nn.Conv2d(inc, 64, kernel_size=(
            7, 7), stride=(2, 2), padding=(3, 3), bias=False)
        self.bn1 = nn.BatchNorm2d(64)
        self.relu = nn.ReLU()
        self.maxpool = nn.MaxPool2d(
            kernel_size=3, stride=2, padding=1, dilation=1, ceil_mode=False)
        self.resblock11 = ResidualBlock2(64, 64)
        self.resblock12 = ResidualBlock2(256, 64)
        self.resblock13 = ResidualBlock2(256, 64)
        self.resblock21 = ResidualBlock2(256, 128)
        self.resblock22 = ResidualBlock2(512, 128)
        self.resblock23 = ResidualBlock2(512, 128)
        self.resblock24 = ResidualBlock2(512, 128)
        self.resblock31 = ResidualBlock2(512, 256)
        self.resblock32 = ResidualBlock2(1024, 256)
        self.resblock33 = ResidualBlock2(1024, 256)
        self.resblock34 = ResidualBlock2(1024, 256)
        self.resblock35 = ResidualBlock2(1024, 256)
        self.resblock36 = ResidualBlock2(1024, 256)
        self.resblock41 = ResidualBlock2(1024, 512)
        self.resblock42 = ResidualBlock2(2048, 512)
        self.resblock43 = ResidualBlock2(2048, 512)
        self.avgpool = nn.AdaptiveAvgPool2d(output_size=(1, 1))

    def forward(self, x):
        '''
        Calculate features from input image x.

        Parameterts
        -----------
        x : torch.tensor
            input image.
        '''
        pr = self.maxpool(self.relu(self.bn1(self.conv1(x))))
        b1 = self.resblock13(self.resblock12(self.resblock11(pr)))
        b2 = self.resblock24(self.resblock23(
            self.resblock22(self.resblock21(b1))))
        b3 = self.resblock36(self.resblock35(self.resblock34(
            self.resblock33(self.resblock32(self.resblock31(b2))))))
        b4 = self.resblock43(self.resblock42(self.resblock41(b3)))
        b5 = self.avgpool(b4)
        return b5

    def out_features(self):
        '''
        Return number of output features, needed for decoders input channels.
        '''
        return self.features

    def name(self):
        '''
        Return name of encoder.
        '''
        return 'resnet50'


class Resnet101Features(nn.Module):
    '''
    Class used to calculate features using resnet101 architecture.
    https://arxiv.org/pdf/1512.03385.pdf

    Methods
    -------
    forward(x)
        calculate features from input image x.

    out_features()
        return number of output features, needed for decoders input channels.

    name()
        return name of encoder.
    '''

    def __init__(self, inc):
        '''
        Create resnet101 encoder layers.

        Parameters
        ----------
        inc : int
            number of input channels [1, 3, 4].
        '''
        super().__init__()
        self.features = 2048
        self.conv1 = nn.Conv2d(inc, 64, kernel_size=(
            7, 7), stride=(2, 2), padding=(3, 3), bias=False)
        self.bn1 = nn.BatchNorm2d(64)
        self.relu = nn.ReLU()
        self.maxpool = nn.MaxPool2d(
            kernel_size=3, stride=2, padding=1, dilation=1, ceil_mode=False)
        self.resblock11 = ResidualBlock2(64, 64)
        self.resblock12 = ResidualBlock2(256, 64)
        self.resblock13 = ResidualBlock2(256, 64)
        self.resblock21 = ResidualBlock2(256, 128)
        self.resblock22 = ResidualBlock2(512, 128)
        self.resblock23 = ResidualBlock2(512, 128)
        self.resblock24 = ResidualBlock2(512, 128)
        self.resblock31 = ResidualBlock2(512, 256)
        self.resblock32 = ResidualBlock2(1024, 256)
        self.resblock33 = ResidualBlock2(1024, 256)
        self.resblock34 = ResidualBlock2(1024, 256)
        self.resblock35 = ResidualBlock2(1024, 256)
        self.resblock36 = ResidualBlock2(1024, 256)
        self.resblock37 = ResidualBlock2(1024, 256)
        self.resblock38 = ResidualBlock2(1024, 256)
        self.resblock39 = ResidualBlock2(1024, 256)
        self.resblock310 = ResidualBlock2(1024, 256)
        self.resblock311 = ResidualBlock2(1024, 256)
        self.resblock312 = ResidualBlock2(1024, 256)
        self.resblock313 = ResidualBlock2(1024, 256)
        self.resblock314 = ResidualBlock2(1024, 256)
        self.resblock315 = ResidualBlock2(1024, 256)
        self.resblock316 = ResidualBlock2(1024, 256)
        self.resblock317 = ResidualBlock2(1024, 256)
        self.resblock318 = ResidualBlock2(1024, 256)
        self.resblock319 = ResidualBlock2(1024, 256)
        self.resblock320 = ResidualBlock2(1024, 256)
        self.resblock321 = ResidualBlock2(1024, 256)
        self.resblock322 = ResidualBlock2(1024, 256)
        self.resblock323 = ResidualBlock2(1024, 256)
        self.resblock41 = ResidualBlock2(1024, 512)
        self.resblock42 = ResidualBlock2(2048, 512)
        self.resblock43 = ResidualBlock2(2048, 512)
        self.avgpool = nn.AdaptiveAvgPool2d(output_size=(1, 1))

    def forward(self, x):
        '''
        Calculate features from input image x.

        Parameterts
        -----------
        x : torch.tensor
            input image.
        '''
        pr = self.maxpool(self.relu(self.bn1(self.conv1(x))))
        b1 = self.resblock13(self.resblock12(self.resblock11(pr)))
        b2 = self.resblock24(self.resblock23(
            self.resblock22(self.resblock21(b1))))
        b3 = self.resblock36(self.resblock35(self.resblock34(
            self.resblock33(self.resblock32(self.resblock31(b2))))))
        b4 = self.resblock312(self.resblock311(self.resblock310(
            self.resblock39(self.resblock38(self.resblock37(b3))))))
        b5 = self.resblock318(self.resblock317(self.resblock316(
            self.resblock315(self.resblock314(self.resblock313(b4))))))
        b6 = self.resblock323(self.resblock322(
            self.resblock321(self.resblock320(self.resblock319(b5)))))
        b7 = self.resblock43(self.resblock42(self.resblock41(b6)))
        b8 = self.avgpool(b7)
        return b8

    def out_features(self):
        '''
        Return number of output features, needed for decoders input channels.
        '''
        return self.features

    def name(self):
        '''
        Return name of encoder.
        '''
        return 'resnet101'


class Resnet152Features(nn.Module):
    '''
    Class used to calculate features using resnet152 architecture.
    https://arxiv.org/pdf/1512.03385.pdf

    Methods
    -------
    forward(x)
        calculate features from input image x.

    out_features()
        return number of output features, needed for decoders input channels.

    name()
        return name of encoder.
    '''

    def __init__(self, inc):
        '''
        Create resnet152 encoder layers.

        Parameters
        ----------
        inc : int
            number of input channels [1, 3, 4].
        '''
        super().__init__()
        self.features = 2048
        self.conv1 = nn.Conv2d(inc, 64, kernel_size=(
            7, 7), stride=(2, 2), padding=(3, 3), bias=False)
        self.bn1 = nn.BatchNorm2d(64)
        self.relu = nn.ReLU()
        self.maxpool = nn.MaxPool2d(
            kernel_size=3, stride=2, padding=1, dilation=1, ceil_mode=False)
        self.resblock11 = ResidualBlock2(64, 64)
        self.resblock12 = ResidualBlock2(256, 64)
        self.resblock13 = ResidualBlock2(256, 64)
        self.resblock21 = ResidualBlock2(256, 128)
        self.resblock22 = ResidualBlock2(512, 128)
        self.resblock23 = ResidualBlock2(512, 128)
        self.resblock24 = ResidualBlock2(512, 128)
        self.resblock25 = ResidualBlock2(512, 128)
        self.resblock26 = ResidualBlock2(512, 128)
        self.resblock27 = ResidualBlock2(512, 128)
        self.resblock28 = ResidualBlock2(512, 128)
        self.resblock31 = ResidualBlock2(512, 256)
        self.resblock32 = ResidualBlock2(1024, 256)
        self.resblock33 = ResidualBlock2(1024, 256)
        self.resblock34 = ResidualBlock2(1024, 256)
        self.resblock35 = ResidualBlock2(1024, 256)
        self.resblock36 = ResidualBlock2(1024, 256)
        self.resblock37 = ResidualBlock2(1024, 256)
        self.resblock38 = ResidualBlock2(1024, 256)
        self.resblock39 = ResidualBlock2(1024, 256)
        self.resblock310 = ResidualBlock2(1024, 256)
        self.resblock311 = ResidualBlock2(1024, 256)
        self.resblock312 = ResidualBlock2(1024, 256)
        self.resblock313 = ResidualBlock2(1024, 256)
        self.resblock314 = ResidualBlock2(1024, 256)
        self.resblock315 = ResidualBlock2(1024, 256)
        self.resblock316 = ResidualBlock2(1024, 256)
        self.resblock317 = ResidualBlock2(1024, 256)
        self.resblock318 = ResidualBlock2(1024, 256)
        self.resblock319 = ResidualBlock2(1024, 256)
        self.resblock320 = ResidualBlock2(1024, 256)
        self.resblock321 = ResidualBlock2(1024, 256)
        self.resblock322 = ResidualBlock2(1024, 256)
        self.resblock323 = ResidualBlock2(1024, 256)
        self.resblock324 = ResidualBlock2(1024, 256)
        self.resblock325 = ResidualBlock2(1024, 256)
        self.resblock326 = ResidualBlock2(1024, 256)
        self.resblock327 = ResidualBlock2(1024, 256)
        self.resblock328 = ResidualBlock2(1024, 256)
        self.resblock329 = ResidualBlock2(1024, 256)
        self.resblock330 = ResidualBlock2(1024, 256)
        self.resblock331 = ResidualBlock2(1024, 256)
        self.resblock332 = ResidualBlock2(1024, 256)
        self.resblock333 = ResidualBlock2(1024, 256)
        self.resblock334 = ResidualBlock2(1024, 256)
        self.resblock335 = ResidualBlock2(1024, 256)
        self.resblock336 = ResidualBlock2(1024, 256)
        self.resblock41 = ResidualBlock2(1024, 512)
        self.resblock42 = ResidualBlock2(2048, 512)
        self.resblock43 = ResidualBlock2(2048, 512)
        self.avgpool = nn.AdaptiveAvgPool2d(output_size=(1, 1))

    def forward(self, x):
        '''
        Calculate features from input image x.

        Parameterts
        -----------
        x : torch.tensor
            input image.
        '''
        pr = self.maxpool(self.relu(self.bn1(self.conv1(x))))
        b1 = self.resblock13(self.resblock12(self.resblock11(pr)))
        b2 = self.resblock28(self.resblock27(self.resblock26(self.resblock25(
            self.resblock24(self.resblock23(
                self.resblock22(self.resblock21(b1))))))))
        b3 = self.resblock36(self.resblock35(self.resblock34(
            self.resblock33(self.resblock32(self.resblock31(b2))))))
        b4 = self.resblock312(self.resblock311(self.resblock310(
            self.resblock39(self.resblock38(self.resblock37(b3))))))
        b5 = self.resblock318(self.resblock317(self.resblock316(
            self.resblock315(self.resblock314(self.resblock313(b4))))))
        b6 = self.resblock324(self.resblock323(self.resblock322(
            self.resblock321(self.resblock320(self.resblock319(b5))))))
        b7 = self.resblock330(self.resblock329(self.resblock328(
            self.resblock327(self.resblock326(self.resblock325(b6))))))
        b8 = self.resblock336(self.resblock335(self.resblock334(
            self.resblock333(self.resblock332(self.resblock331(b7))))))
        b9 = self.resblock43(self.resblock42(self.resblock41(b8)))
        b10 = self.avgpool(b9)
        return b10

    def out_features(self):
        '''
        Return number of output features, needed for decoders input channels.
        '''
        return self.features

    def name(self):
        '''
        Return name of encoder.
        '''
        return 'resnet152'



from .accuracy import accuracy
from .precision import precision
from .recall import recall
from .f1_score import f1_score


def get_metrics(metrics, nc=None):
    '''
    Return list of metrics by names.

    Parameterts
    -----------
    metrics : list[str]
        list of metrics names.

    nc : int
        number classes.
    '''
    metrics_fn = []
    for metric in metrics:
        if metric == 'accuracy':
            acc = accuracy()
            metrics_fn.append(acc)
        elif metric == 'precision':
            prec = precision(nc)
            metrics_fn.append(prec)
        elif metric == 'recall':
            rec = recall(nc)
            metrics_fn.append(rec)
        elif metric == 'f1_score':
            f1 = f1_score(nc)
            metrics_fn.append(f1)
        else:
            raise ValueError(f'no metric {metric}')
    return metrics_fn



import torchvision
import torch


def gtsrb(tr_part, va_part, te_part, transform):
    '''
    Return splited into train, val, test parts gtsrb dataset.

    Parameterts
    -----------
    tr_part : float
        percent of train data.

    va_part : float
        percent of valid data.

    te_part : float
        percent of test data.

    transform : torchvision.transforms
        torchvision transforms object with augmentations.
    '''
    dataset = torchvision.datasets.GTSRB(root='.',
                                         download=True,
                                         transform=transform)

    tr_part = int(tr_part * len(dataset))
    va_part = int(va_part * len(dataset))
    te_part = int(te_part * len(dataset))

    if tr_part + va_part + te_part != len(dataset):
        sizes = [tr_part,
                 va_part,
                 te_part,
                 len(dataset) - tr_part - va_part - te_part]
    else:
        sizes = [tr_part, va_part, te_part]

    datasets = torch.utils.data.random_split(dataset, sizes)
    train_dataset = datasets[0]
    valid_dataset = datasets[1]
    test_dataset = datasets[2]

    return train_dataset, valid_dataset, test_dataset



import torch
from torch import nn


class ConvBlock(nn.Module):
    '''
    Conv block class.

    Methods
    -------
    forward(x)
        calculate features from input image x.
    '''

    def __init__(self, inc, outc, **kwargs):
        '''
        Create convolution block layers.

        Parameters
        ----------
        inc : int
            number of input channels.

        outc : int
            number of output channels.
        '''
        super().__init__()
        self.conv = nn.Conv2d(inc, outc, bias=False, **kwargs)
        self.bn = nn.BatchNorm2d(outc, eps=0.001)
        self.relu = nn.ReLU()

    def forward(self, x):
        '''
        Calculate features from input image x.

        Parameterts
        -----------
        x : torch.tensor
            input image.
        '''
        return self.relu(self.bn(self.conv(x)))


class InceptionBlock(nn.Module):
    '''
    Inception block class.

    Methods
    -------
    forward(x)
        calculate features from input image x.
    '''

    def __init__(self, inc, c1x1, cr3x3, c3x3, cr5x5, c5x5, pool):
        '''
        Create inception block layers.

        Parameters
        ----------
        inc : int
            number of input channels.

        c1x1 : int
            number of channels for convolution 1x1 in single branch.

        cr3x3 : int
            number of channels for convolution 1x1 in branch reduce 3x3.

        c3x3 : int
            number of channels for convolution 3x3 in branch reduce 3x3.

        cr5x5 : int
            number of channels for convolution 1x1 in branch reduce 5x5.

        c5x5 : int
            number of channels for convolution 5x5 in branch reduce 5x5.

        pool : int
            number of channels for convolution 1x1 in branch pooling.
        '''
        super().__init__()
        self.branch1 = ConvBlock(inc, c1x1, kernel_size=(1, 1))
        self.branch2 = nn.Sequential(
            ConvBlock(inc, cr3x3, kernel_size=(1, 1)),
            ConvBlock(cr3x3, c3x3, kernel_size=(3, 3), padding=(1, 1))
        )
        self.branch3 = nn.Sequential(
            ConvBlock(inc, cr5x5, kernel_size=(1, 1)),
            ConvBlock(cr5x5, c5x5, kernel_size=(5, 5), padding=(2, 2))
        )
        self.branch4 = nn.Sequential(
            nn.MaxPool2d(kernel_size=3, stride=1, padding=1),
            ConvBlock(inc, pool, kernel_size=(1, 1))
        )

    def forward(self, x):
        '''
        Calculate features from input image x.

        Parameterts
        -----------
        x : torch.tensor
            input image.
        '''
        br1 = self.branch1(x)
        br2 = self.branch2(x)
        br3 = self.branch3(x)
        br4 = self.branch4(x)
        cat_br = torch.cat((br1, br2, br3, br4), dim=1)
        return cat_br


class GoognetNoAuxFeatures(nn.Module):
    '''
    Class used to calculate features using googlenet architecture.
    https://arxiv.org/pdf/1409.4842.pdf

    Methods
    -------
    forward(x)
        calculate features from input image x.

    out_features()
        return number of output features, needed for decoders input channels.

    name()
        return name of encoder.
    '''

    def __init__(self, inc):
        '''
        Create googlenet encoder layers.

        Parameters
        ----------
        inc : int
            number of input channels [1, 3, 4].
        '''
        super().__init__()
        self.features = 1024

        self.conv1 = ConvBlock(inc,
                               64,
                               kernel_size=(7, 7),
                               stride=(2, 2),
                               padding=(3, 3))
        self.maxpool1 = nn.MaxPool2d(kernel_size=3, stride=2)
        self.conv2 = ConvBlock(64, 64, kernel_size=(1, 1))
        self.conv3 = ConvBlock(64, 192, kernel_size=(3, 3), padding=(1, 1))
        self.maxpool2 = nn.MaxPool2d(kernel_size=3, stride=2)

        self.inception3a = InceptionBlock(192, 64, 96, 128, 16, 32, 32)
        self.inception3b = InceptionBlock(256, 128, 128, 192, 32, 96, 64)
        self.maxpool3 = nn.MaxPool2d(kernel_size=3, stride=2)

        self.inception4a = InceptionBlock(480, 192, 96, 208, 16, 48, 64)
        self.inception4b = InceptionBlock(512, 160, 112, 224, 24, 64, 64)
        self.inception4c = InceptionBlock(512, 128, 128, 256, 24, 64, 64)
        self.inception4d = InceptionBlock(512, 112, 144, 288, 32, 64, 64)
        self.inception4e = InceptionBlock(528, 256, 160, 320, 32, 128, 128)
        self.maxpool4 = nn.MaxPool2d(kernel_size=3, stride=2)

        self.inception5a = InceptionBlock(832, 256, 160, 320, 32, 128, 128)
        self.inception5b = InceptionBlock(832, 384, 192, 384, 48, 128, 128)

        self.avgpool = nn.AdaptiveAvgPool2d((1, 1))

    def forward(self, x):
        '''
        Calculate features from input image x.

        Parameterts
        -----------
        x : torch.tensor
            input image.
        '''
        pr = self.maxpool2(self.conv3(
            self.conv2(self.maxpool1(self.conv1(x)))))
        inc3 = self.maxpool3(self.inception3b(self.inception3a(pr)))
        inc4 = self.maxpool4(self.inception4e(self.inception4d(
            self.inception4c((self.inception4b(self.inception4a(inc3)))))))
        inc5 = self.avgpool(self.inception5b(self.inception5a(inc4)))
        return inc5

    def out_features(self):
        '''
        Return number of output features, needed for decoders input channels.
        '''
        return self.features

    def name(self):
        '''
        Return name of encoder.
        '''
        return 'googlenetnoaux'



class accuracy():
    '''
    Class used to calculate accuracy metric.

    Methods
    -------
    calc(preds, labels)
        calculate metric.

    name()
        return metric name.
    '''
    def calc(self, preds, labels):
        '''
        Calculate accuracy metric.

        Parameterts
        -----------
        preds : torch.tensor
            model predictions.

        labels : torch.tensor
            ground-truth labels.
        '''
        np = 1
        shapes = preds.shape

        if len(shapes) == len(labels.shape) and shapes[1] != labels.shape[1]:
            preds = preds.argmax(dim=1).unsqueeze(1)
        elif len(shapes) > len(labels.shape):
            preds = preds.argmax(dim=1)
        if len(labels.shape) == 2:
            preds = preds.argmax(dim=1).float()
            labels = labels.argmax(dim=1).float()

        for shape in labels.shape:
            np *= shape

        acc = (preds == labels).sum() / np
        return acc

    def name(self):
        '''
        Return metric name.
        '''
        return 'accuracy'



from .adam import adam, adamw, adamax, radam, nadam


def get_optimizer(name,
                  model,
                  lr=None,
                  betas=None,
                  eps=None,
                  weight_decay=None):
    '''
    Return optimizer by name.

    Parameterts
    -----------
    name : str
        optimizer name ['adam', 'adamw', 'adamax', 'radam', 'nadam'].

    model : Any
        pytorch model.

    lr : float, optional
        learning rate.

    betas : tuple(float, float), optional
        betas.

    eps : float, optional
        eps.

    weight_decay : float, optional
        l2 regularization.
    '''
    if name == 'adam':
        optimizer = adam(model.parameters(), lr, betas, eps, weight_decay)
    elif name == 'adamw':
        optimizer = adamw(model.parameters(), lr, betas, eps, weight_decay)
    elif name == 'adamax':
        optimizer = adamax(model.parameters(), lr, betas, eps, weight_decay)
    elif name == 'radam':
        optimizer = radam(model.parameters(), lr, betas, eps, weight_decay)
    elif name == 'nadam':
        optimizer = nadam(model.parameters(), lr, betas, eps, weight_decay)
    else:
        raise ValueError(f'no optimizer {name}')
    return optimizer



from torch import nn


class ConvBlock(nn.Module):
    '''
    Conv block class.

    Methods
    -------
    forward(x)
        calculate features from input image x.
    '''

    def __init__(self, inc, outc, **kwargs):
        '''
        Create convolution block layers.

        Parameters
        ----------
        inc : int
            number of input channels.

        outc : int
            number of output channels.
        '''
        super().__init__()
        self.conv = nn.Conv2d(inc, outc, bias=False, **kwargs)
        self.bn = nn.BatchNorm2d(outc)
        self.relu = nn.ReLU()

    def forward(self, x):
        '''
        Calculate features from input image x.

        Parameterts
        -----------
        x : torch.tensor
            input image.
        '''
        return self.relu(self.bn(self.conv(x)))


class DepthWiseSepConvBlock(nn.Module):
    '''
    Depth wise separable convolution class.

    Methods
    -------
    forward(x)
        calculate features from input image x.
    '''

    def __init__(self, inc, outc, **kwargs):
        '''
        Create depth wise separable convolution layers.

        Parameters
        ----------
        inc : int
            number of input channels.

        outc : int
            number of output channels.
        '''
        super().__init__()
        self.depthwise = nn.Conv2d(inc,
                                   inc,
                                   kernel_size=(3, 3),
                                   padding=(1, 1),
                                   groups=inc,
                                   bias=False,
                                   **kwargs)
        self.bn1 = nn.BatchNorm2d(inc)
        self.pointwise = nn.Conv2d(inc, outc, kernel_size=(1, 1), bias=False)
        self.bn2 = nn.BatchNorm2d(outc)
        self.relu = nn.ReLU()

    def forward(self, x):
        '''
        Calculate features from input image x.

        Parameterts
        -----------
        x : torch.tensor
            input image.
        '''
        depth = self.relu(self.bn1(self.depthwise(x)))
        point = self.relu(self.bn2(self.pointwise(depth)))
        return point


class MobilenetFeatures(nn.Module):
    '''
    Class used to calculate features using mobilenet architecture.
    https://arxiv.org/pdf/1704.04861.pdf

    Methods
    -------
    forward(x)
        calculate features from input image x.

    out_features()
        return number of output features, needed for decoders input channels.

    name()
        return name of encoder.
    '''

    def __init__(self, inc):
        '''
        Create mobilenet encoder layers.

        Parameters
        ----------
        inc : int
            number of input channels [1, 3, 4].
        '''
        super().__init__()
        self.features = 1024
        self.conv = ConvBlock(inc,
                              32,
                              kernel_size=(3, 3),
                              stride=(2, 2),
                              padding=(1, 1))
        self.convdws1 = DepthWiseSepConvBlock(32, 64, stride=(1, 1))
        self.convdws2 = DepthWiseSepConvBlock(64, 128, stride=(2, 2))
        self.convdws3 = DepthWiseSepConvBlock(128, 128, stride=(1, 1))
        self.convdws4 = DepthWiseSepConvBlock(128, 256, stride=(2, 2))
        self.convdws5 = DepthWiseSepConvBlock(256, 256, stride=(1, 1))
        self.convdws6 = DepthWiseSepConvBlock(256, 512, stride=(2, 2))
        self.convdws7 = DepthWiseSepConvBlock(512, 512, stride=(1, 1))
        self.convdws8 = DepthWiseSepConvBlock(512, 512, stride=(1, 1))
        self.convdws9 = DepthWiseSepConvBlock(512, 512, stride=(1, 1))
        self.convdws10 = DepthWiseSepConvBlock(512, 512, stride=(1, 1))
        self.convdws11 = DepthWiseSepConvBlock(512, 512, stride=(1, 1))
        self.convdws12 = DepthWiseSepConvBlock(512, 1024, stride=(2, 2))
        self.convdws13 = DepthWiseSepConvBlock(1024, 1024, stride=(1, 1))
        self.avgpool = nn.AdaptiveAvgPool2d(output_size=(1, 1))

    def forward(self, x):
        '''
        Calculate features from input image x.

        Parameterts
        -----------
        x : torch.tensor
            input image.
        '''
        b1 = self.convdws4(self.convdws3(
            self.convdws2(self.convdws1(self.conv(x)))))
        b2 = self.convdws9(self.convdws8(
            self.convdws7(self.convdws6(self.convdws5(b1)))))
        b3 = self.avgpool(self.convdws13(
            self.convdws12(self.convdws11(self.convdws10(b2)))))
        return b3

    def out_features(self):
        '''
        Return number of output features, needed for decoders input channels.
        '''
        return self.features

    def name(self):
        '''
        Return name of encoder.
        '''
        return 'mobilenet'



from .steplr import steplr, multisteplr
from .custom import polylr


def get_scheduler(name,
                  optim,
                  step=None,
                  gamma=None,
                  milestones=None,
                  max_decay_steps=None,
                  end_lr=None,
                  power=None):
    '''
    Return scheduler by name.

    Parameterts
    -----------
    name : str
        scheduler name ['steplr', 'multisteplr']

    optim : Optimizer
        optimizer object.

    step : int, optional
        every step epochs do scheduler step.

    gamma : float, optional
        lr will be multiplied to this value each step epochs.

    milestones : list[int], optional
        checkpoints for lr * gamma.

    max_decay_steps : int
        maximum number of decay steps.

    end_lr : float
        minimal lr.

    power : float
        power value.
    '''
    if name == 'steplr':
        scheduler = steplr(optim, step, gamma)
    elif name == 'multisteplr':
        scheduler = multisteplr(optim, milestones, gamma)
    elif name == 'polylr':
        scheduler = polylr(optim, max_decay_steps, end_lr, power)
    else:
        raise ValueError(f'no scheduler {name}')
    return scheduler



from torch.optim import Adam, NAdam, RAdam, AdamW, Adamax


def adam(parameters, lr, betas, eps, weight_decay):
    '''
    Return Adam optimizer object.
    https://pytorch.org/docs/stable/generated/torch.optim.Adam.html#torch.optim.Adam

    Parameterts
    -----------
    parameters : Any
        model.parameters()

    lr : float
        learning rate

    betas : tuple(float, float)
        betas

    eps : float
        eps

    weight_decay : float
        l2 regularization
    '''
    return Adam(parameters, lr=lr,
                betas=betas,
                eps=eps,
                weight_decay=weight_decay)


def nadam(parameters, lr, betas, eps, weight_decay):
    '''
    Return NAdam optimizer object.
    https://pytorch.org/docs/stable/generated/torch.optim.NAdam.html#torch.optim.NAdam

    Parameterts
    -----------
    parameters : Any
        model.parameters().

    lr : float
        learning rate.

    betas : tuple(float, float)
        betas.

    eps : float
        eps.

    weight_decay : float
        l2 regularization.
    '''
    return NAdam(parameters,
                 lr=lr,
                 betas=betas,
                 eps=eps,
                 weight_decay=weight_decay)


def radam(parameters, lr, betas, eps, weight_decay):
    '''
    Return RAdam optimizer object.
    https://pytorch.org/docs/stable/generated/torch.optim.RAdam.html#torch.optim.RAdam

    Parameterts
    -----------
    parameters : Any
        model.parameters().

    lr : float
        learning rate.

    betas : tuple(float, float)
        betas.

    eps : float
        eps.

    weight_decay : float
        l2 regularization.
    '''
    return RAdam(parameters,
                 lr=lr,
                 betas=betas,
                 eps=eps,
                 weight_decay=weight_decay)


def adamw(parameters, lr, betas, eps, weight_decay):
    '''
    Return AdamW optimizer object.
    https://pytorch.org/docs/stable/generated/torch.optim.AdamW.html#torch.optim.AdamW

    Parameterts
    -----------
    parameters : Any
        model.parameters().

    lr : float
        learning rate.

    betas : tuple(float, float)
        betas.

    eps : float
        eps.

    weight_decay : float
        l2 regularization.
    '''
    return AdamW(parameters,
                 lr=lr,
                 betas=betas,
                 eps=eps,
                 weight_decay=weight_decay)


def adamax(parameters, lr, betas, eps, weight_decay):
    '''
    Return Adamax optimizer object.
    https://pytorch.org/docs/stable/generated/torch.optim.Adamax.html#torch.optim.Adamax

    Parameterts
    -----------
    parameters : Any
        model.parameters().

    lr : float
        learning rate.

    betas : tuple(float, float)
        betas.

    eps : float
        eps.

    weight_decay : float
        l2 regularization.
    '''
    return Adamax(parameters,
                  lr=lr,
                  betas=betas,
                  eps=eps,
                  weight_decay=weight_decay)



from sklearn.metrics import recall_score
import torch


class recall():
    '''
    Class used to calculate recall metric.

    Attributes
    ----------
    nc : int
        classes number.

    Methods
    -------
    calc(preds, labels)
        calculate metric.

    name()
        return metric name.
    '''

    def __init__(self, nc):
        '''
        Parameters
        ----------
        nc : int
            classes number.
        '''
        self.nc = nc

    def calc(self, preds, labels):
        '''
        Calculate recall metric.

        Parameterts
        -----------
        preds : torch.tensor
            model predictions.

        labels : torch.tensor
            ground-truth labels.
        '''
        shapes = preds.shape

        if len(shapes) == len(labels.shape) and shapes[1] != labels.shape[1]:
            preds = preds.argmax(dim=1).unsqueeze(1)
        elif len(shapes) > len(labels.shape):
            preds = preds.argmax(dim=1)

        if len(labels.shape) == 2:
            preds = preds.argmax(dim=1).float()
            labels = labels.argmax(dim=1).float()

        return torch.tensor(recall_score(labels.detach().cpu(),
                                         preds.detach().cpu(),
                                         average='micro'))

    def name(self):
        '''
        Return metric name.
        '''
        return 'recall'



from torch import nn


class AlexnetFeatures(nn.Module):
    '''
    Class used to calculate features using alexnet architecture.
    https://proceedings.neurips.cc/paper/2012/file/c399862d3b9d6b76c8436e924a68c45b-Paper.pdf

    Methods
    -------
    forward(x)
        calculate features from input image x.

    out_features()
        return number of output features, needed for decoders input channels.

    name()
        return name of encoder.
    '''

    def __init__(self, inc):
        '''
        Create alexnet encoder layers.

        Parameters
        ----------
        inc : int
            number of input channels [1, 3, 4].
        '''
        super().__init__()
        self.features = 256
        self.conv1 = nn.Conv2d(inc,
                               64,
                               kernel_size=(11, 11),
                               stride=(4, 4),
                               padding=(2, 2))
        self.conv2 = nn.Conv2d(64,
                               192,
                               kernel_size=(5, 5),
                               stride=(1, 1),
                               padding=(2, 2))
        self.conv3 = nn.Conv2d(192,
                               384,
                               kernel_size=(3, 3),
                               stride=(1, 1),
                               padding=(1, 1))
        self.conv4 = nn.Conv2d(384,
                               256,
                               kernel_size=(3, 3),
                               stride=(1, 1),
                               padding=(1, 1))
        self.conv5 = nn.Conv2d(256,
                               256,
                               kernel_size=(3, 3),
                               stride=(1, 1),
                               padding=(1, 1))
        self.maxpool = nn.MaxPool2d(kernel_size=3,
                                    stride=2,
                                    padding=0,
                                    dilation=1,
                                    ceil_mode=False)
        self.avgpool = nn.AdaptiveAvgPool2d(output_size=(1, 1))
        self.relu = nn.ReLU()

    def forward(self, x):
        '''
        Calculate features from input image x.

        Parameterts
        -----------
        x : torch.tensor
            input image.
        '''
        b1 = self.maxpool(self.relu(self.conv1(x)))
        b2 = self.maxpool(self.relu(self.conv2(b1)))
        b3 = self.relu(self.conv3(b2))
        b4 = self.relu(self.conv4(b3))
        b5 = self.avgpool(self.relu(self.conv5(b4)))
        return b5

    def out_features(self):
        '''
        Return number of output features, needed for decoders input channels.
        '''
        return self.features

    def name(self):
        '''
        Return name of encoder.
        '''
        return 'alexnet'



from torch.nn import CrossEntropyLoss
from torch import nn
import torch


def celoss():
    '''
    Return cross-entropy loss object.
    https://pytorch.org/docs/stable/generated/torch.nn.CrossEntropyLoss.html
    '''
    return CrossEntropyLoss()


def customceloss():
    '''
    Return custom cross-entropy loss object from CustomCrossEntropyLoss class.
    https://pytorch.org/docs/stable/generated/torch.nn.CrossEntropyLoss.html
    '''
    return CustomCrossEntropyLoss()


class CustomCrossEntropyLoss(nn.Module):
    '''
    Class used to calculate cross-entropy loss.

    Attributes
    ----------
    weight : float, list[float]
        weight for loss value.

    reduction : str
        type of reduction ['mean', 'sum'].

    Methods
    -------
    forward(inputs, targets)
        calculate cross-entropy loss between inputs and targets.
    '''

    def __init__(self, weight=1.0, reduction='mean'):
        '''
        Parameters
        ----------
        weight : float, list[float], optional
            weight for loss value.

        reduction : str, optional
            type of reduction ['mean', 'sum'].
        '''
        super().__init__()
        self.weight = weight
        self.reduction = reduction

    def forward(self, inputs, targets):
        '''
        Calculate cross-entropy loss between inputs and targets.

        Parameterts
        -----------
        inputs : torch.tensor
            model predictions.

        targets : torch.tensor
            ground-truth labels.
        '''
        inputs = inputs.view(inputs.shape[0], -1)
        exps = torch.exp(torch.gather(inputs, 1, targets.view(-1, 1)))
        expinputs = torch.exp(inputs).sum()
        logs = torch.clamp(torch.log2(exps / expinputs), min=-100.0)
        loss = -self.weight * logs
        if self.reduction == 'mean':
            return loss.mean()
        elif self.reduction == 'sum':
            return loss.sum()



from torch import nn


class LenetDecoder(nn.Module):
    '''
    Class used to decode features using lenet architecture.
    http://vision.stanford.edu/cs598_spring07/papers/Lecun98.pdf

    Methods
    -------
    forward(x)
        calculate output from input features x.
    '''
    def __init__(self, inf, nc):
        '''
        Create lenet decoder layers.

        Parameters
        ----------
        inf : int
            number of input features.

        nc : int
            number output classes.
        '''
        super().__init__()
        self.fc1 = nn.Linear(inf, 120, bias=True)
        self.fc2 = nn.Linear(120, 84, bias=True)
        self.fc3 = nn.Linear(84, nc, bias=True)
        self.drop = nn.Dropout(p=0.5)
        self.tanh = nn.Tanh()
        self.softmax = nn.Softmax(dim=1)

    def forward(self, x):
        '''
        Calculate output from input features x.

        Parameterts
        -----------
        x : torch.tensor
            input features.
        '''
        b1 = self.tanh(self.fc1(self.drop(x.view(x.shape[0], -1))))
        b2 = self.tanh(self.fc2(self.drop(b1)))
        b3 = self.softmax(self.fc3(b2))
        return b3



from torch import nn


class LinearDecoder(nn.Module):
    '''
    Class used to decode features using one linear layer.

    Methods
    -------
    forward(x)
        calculate output from input features x.
    '''
    def __init__(self, inf, nc):
        '''
        Create one linear decoder layer.

        Parameters
        ----------
        inf : int
            number of input features.

        nc : int
            number output classes.
        '''
        super().__init__()
        self.fc = nn.Linear(inf, nc)
        self.softmax = nn.Softmax(dim=1)

    def forward(self, x):
        '''
        Calculate output from input features x.

        Parameterts
        -----------
        x : torch.tensor
            input features.
        '''
        return self.softmax(self.fc(x.view(x.shape[0], -1)))



import torch
from torch import nn
from .alexnet import AlexnetDecoder
from .lenet import LenetDecoder
from .linear import LinearDecoder


class Model(nn.Module):
    '''
    Class used to combine encoder and decoder models.
    One encoder - one decoder.

    Attributes
    ----------
    encoder : Any
        encoder model.

    decoder : Any
        decoder model.

    inf : int
        number of encoder output features,
        used as number input features for decoder.

    Methods
    -------
    forward(x)
        encode input image x into features, decode features into output.
    '''

    def __init__(self, name, encoder, nc, device):
        '''
        Create model by combine one encoder and one decoder.

        Parameters
        ----------
        name : str
            decoder name.

        encoder : Any
            encoder model.

        nc : int
            classes number.

        device : str
            device ['cpu', 'cuda']
        '''
        super().__init__()
        self.encoder = encoder.to(device)
        self.inf = encoder.out_features()
        if name == 'alexnet':
            self.decoder = AlexnetDecoder(self.inf, nc).to(device)
        elif name == 'lenet':
            self.decoder = LenetDecoder(self.inf, nc).to(device)
        elif name == 'linear':
            self.decoder = LinearDecoder(self.inf, nc).to(device)
        else:
            raise ValueError(f'no decoder {name}')

    def forward(self, x):
        '''
        Encode input image x into features, decode features into output.

        Parameterts
        -----------
        x : torch.tensor
            input image.
        '''
        features = self.encoder(x)
        return self.decoder(features)


class MultiDecModel(nn.Module):
    '''
    Class used to combine encoder and decoder models.
    One encoder - nc decoders.

    Attributes
    ----------
    encoder : Any
        encoder model.

    decoders : list[Any]
        list with decoder models.

    inf : int
        number of encoder output features,
        used as number input features for decoder.

    nc : int
        classes number.

    Methods
    -------
    forward(x)
        encode input image x into features, decode features into output.
    '''

    def __init__(self, name, encoder, nc, device):
        '''
        Create model by combine one encoder and nc decoders.

        Parameters
        ----------
        name : str
            decoder name.

        encoder : Any
            encoder model.

        nc : int
            classes number.

        device : str
            device ['cpu', 'cuda']
        '''
        super().__init__()
        self.nc = nc
        self.encoder = encoder
        self.inf = encoder.out_features()
        self.decoders = []
        for i in range(nc):
            if name[i] == 'alexnet':
                self.decoders.append(AlexnetDecoder(self.inf, 1).to(device))
            elif name[i] == 'lenet':
                self.decoders.append(LenetDecoder(self.inf, 1).to(device))
            elif name == 'linear':
                self.decoders.append(LinearDecoder(self.inf, 1).to(device))
            else:
                raise ValueError(f'no decoder {name}')

    def forward(self, x):
        '''
        Encode input image x into features, decode features into output.

        Parameterts
        -----------
        x : torch.tensor
            input image.
        '''
        features = self.encoder(x)
        res = torch.cat([self.decoders[i](features)
                        for i in range(self.nc)], dim=1)
        return res



from opennn_pytorch.algo import train, test, prediction
from opennn_pytorch.optimizers import get_optimizer
from opennn_pytorch.schedulers import get_scheduler
from opennn_pytorch.datasets import get_dataset
from opennn_pytorch.metrics import get_metrics
from opennn_pytorch.losses import get_loss
from opennn_pytorch.encoders import get_encoder
from opennn_pytorch.decoders import get_decoder

import yaml
import random
import numpy as np
import torch
from torchvision import transforms
import os
import wandb


def parse_yaml(config):
    '''
    Parameterts
    -----------
    config : str
        path to .yaml config file.
    '''
    with open(config, 'r') as stream:
        data = yaml.safe_load(stream)
    return data


def transforms_lst(transform_config):
    '''
    Transforms dict into list of torchvision.transforms.

    Parameterts
    -----------
    transform_config : dict[str, Any]
        dict from .yaml config file.
    '''
    lst = []
    for el in transform_config.keys():
        if el == 'tensor' and transform_config['tensor']:
            lst.append(transforms.ToTensor())
        elif el == 'resize':
            size = int(transform_config[el])
            lst.append(transforms.Resize((size, size)))
        elif el == 'normalize':
            means_stds = transform_config[el]
            means = list(map(float, means_stds[0]))
            stds = list(map(float, means_stds[1]))
            lst.append(transforms.Normalize(means, stds))
        else:
            raise ValueError(f'no augmentation {el}: {transform_config[el]}')
    return lst


def run(yaml):
    '''
    Parse .yaml config and transforms .yaml config
    and generate full train/test pipeline.

    Parameterts
    -----------
    yaml : str
        main .yaml config with all basic parameters.

    Config Attributes
    -----------------
    model : structure
        architecture : structure
            encoder : str
                [lenet, alexnet, googlenet, resnet18, resnet34, resnet50,
                 resnet101, resnet152, mobilenet, vgg11, vgg16, vgg19]

            decoder : str, optional
                [lenet, alexnet, linear]

            multidecoder : list[decoder], optional

        features : structure
            in_channels : int

            number_classes : int

        checkpoint : str, optional
            if specify this checkpoint will be loaded into model.

    algorithm : structure
        name : str
            [train, test]

        device : str
            [cpu, cuda]

        epochs : int

        seed : int

    dataset : structure
        name : str
            [mnist, fashion_mnist, cifar10, cifar100, gtsrb, custom]

        images : str, optional
            specify path to folder with images, only for custom dataset.

        annotation : str, optional
            specify path to yaml file with labels - images specification,
            only for custom dataset.

        batch_size : int

        sizes : structure
            train_size : float
                [0.0 - 1.0]

            valid_size : float
                [0.0 - 1.0]

            test_size : float
                [0.0 - 1.0]

        transform : str
            auxiliary .yaml config with sequential transforms
            for image preprocessing.

    save : structure
        logs : structure
            path : str

        checkpoints : structure
            path : str

            save_every : int

    optimizer : structure
        nane : str
            [adam, adamw, adamax, radam, nadam]

        params : structure
            learning_rate : float

            betas : list[float, float], optional

            eps : float, optional

            weight_decay : float, optional

    scheduler : structure
        name : str
            [steplr, multisteplr, polylr]

        params : structure
            step : int, optional

            gamma : float, optional

            milestones : list[int], optional

            max_decay_steps : int, optional

            end_lr : float, optional

            power : float, optional

    loss_function : str
        [ce, custom_ce, bce, custom_bce, bce_logits, custom_bce_logits, mse,
         custom_mse, mae, custom_mae]

    metrics : list[str]
        str : [accuracy, precision, recall, f1_score]
        len(str) : [1 - 4]

    class_names : list[str], optional
        if specify 10 random images will be vizualized.

    wandb: structure, optional
        project_name : str
            name of wandb project.

        run_name : str
            name of run in wandb project.

        metrics : list[str]
            metric names which will be logged by wandb.

    Transform Config Attributes
    ---------------------------
    tensor : bool

    resize : int

    normalize : list[list[float], list[float]]
        [[means], [stds]] - where means and stds count
        = number of input channels
    '''
    torch.cuda.empty_cache()

    config = parse_yaml(yaml)
    transform_config = parse_yaml(config['dataset']['transform'])
    transform_lst = transforms_lst(transform_config)
    transform = transforms.Compose(transform_lst)

    encoder_name = config['model']['architecture']['encoder']

    if 'decoder' in config['model']['architecture'].keys():
        decoder_name = config['model']['architecture']['decoder']
        decoder_mode = None
    else:
        decoder_name = config['model']['architecture']['multidecoder']
        decoder_mode = 'multidecoder'

    inc = int(config['model']['features']['in_channels'])
    nc = int(config['model']['features']['number_classes'])

    device = config['algorithm']['device']

    algorithm = config['algorithm']['name']

    dataset = config['dataset']['name']
    if dataset == 'custom':
        datafiles = (config['dataset']['images'],
                     config['dataset']['annotation'])
    else:
        datafiles = None

    train_part = float(config['dataset']['sizes']['train_size'])
    valid_part = float(config['dataset']['sizes']['valid_size'])
    test_part = float(config['dataset']['sizes']['test_size'])
    seed = int(config['algorithm']['seed'])

    bs = int(config['dataset']['batch_size'])
    epochs = int(config['algorithm']['epochs'])

    logs = config['save']['logs']['path']
    if not os.path.isdir(logs):
        cwd = os.getcwd().replace('\\', '/')
        os.mkdir(os.path.join(cwd, logs), 0o777)

    loss_fn = config['loss_function']
    metrics = config['metrics']

    checkpoints = config['save']['checkpoints']['path']
    if not os.path.isdir(checkpoints):
        cwd = os.getcwd().replace('\\', '/')
        os.mkdir(os.path.join(cwd, checkpoints), 0o777)

    if algorithm == 'test' and 'class_names' in config.keys():
        names = config['class_names']
        pred = True
    else:
        pred = False

    if 'checkpoint' in config['model'].keys():
        checkpoint = config['model']['checkpoint']
    else:
        checkpoint = None

    se = int(config['save']['checkpoints']['save_every'])
    lr = float(config['optimizer']['params']['learning_rate'])

    if 'weight_decay' in config['optimizer']['params'].keys():
        wd = float(config['optimizer']['params']['weight_decay'])
    else:
        wd = 0.0

    if 'eps' in config['optimizer']['params'].keys():
        opt_eps = float(config['optimizer']['params']['eps'])
    else:
        opt_eps = 1e-8

    if 'betas' in config.keys():
        betas = tuple(list(map(float, config['optimizer']['params']['betas'])))
    else:
        betas = (0.9, 0.999)

    optim = config['optimizer']['name']
    sched = config['scheduler']['name']

    if 'step' in config['scheduler']['params'].keys():
        step = int(config['scheduler']['params']['step'])
    else:
        step = 10

    if 'gamma' in config['scheduler']['params'].keys():
        gamma = float(config['scheduler']['params']['gamma'])
    else:
        gamma = 0.1

    if 'milestones' in config['scheduler']['params'].keys():
        milestones = list(
            map(int, config['scheduler']['params']['milestones']))
    else:
        milestones = [10, 30, 70, 150]

    if 'max_decay_steps' in config['scheduler']['params'].keys():
        mdsteps = int(config['scheduler']['params']['max_decay_steps'])
    else:
        mdsteps = 100

    if 'end_lr' in config.keys():
        end_lr = float(config['scheduler']['params']['end_lr'])
    else:
        end_lr = 0.00001

    if 'power' in config['scheduler']['params'].keys():
        power = float(config['scheduler']['params']['power'])
    else:
        power = 1.0

    wandb_flag = False
    wandb_metrics = None

    if 'wandb' in config.keys():
        wandb_flag = True
        wandb_dct = {}

        wandb_dct['algorithm'] = algorithm
        wandb_dct['seed'] = seed
        wandb_dct['encoder'] = encoder_name
        wandb_dct['decoder'] = decoder_name
        wdm = 'single' if decoder_mode is None else 'multi'
        wandb_dct['decoder mode'] = wdm
        wandb_dct['input channels'] = inc
        wandb_dct['number classes'] = nc
        wandb_dct['device'] = device
        wandb_dct['initial learning rate'] = lr
        wandb_dct['weight decay'] = wd
        wandb_dct['batch size'] = bs
        wandb_dct['dataset name'] = config['dataset']['name']
        wandb_dct['train size'] = train_part
        wandb_dct['valid size'] = valid_part
        wandb_dct['test size'] = test_part
        wandb_dct['loss function'] = config['loss_function']
        wandb_dct['optimizer'] = config['optimizer']['name']
        wandb_dct['scheduler'] = config['scheduler']['name']

        wandb.init(project=config['wandb']['project_name'],
                   name=config['wandb']['run_name'],
                   config=wandb_dct)
        wandb_metrics = config['wandb']['metrics']

    np.random.seed(seed)
    torch.manual_seed(seed)

    if device != 'cpu' and device != 'cuda':
        raise ValueError(f'no device {device}')

    encoder = get_encoder(encoder_name, inc).to(device)
    model = get_decoder(decoder_name,
                        encoder,
                        nc,
                        decoder_mode,
                        device).to(device)

    if checkpoint is not None:
        state_dict = torch.load(checkpoint)
        model.load_state_dict(state_dict)

    train_data, valid_data, test_data = get_dataset(dataset,
                                                    train_part,
                                                    valid_part,
                                                    test_part,
                                                    transform,
                                                    datafiles)

    optimizer = get_optimizer(optim,
                              model,
                              lr=lr,
                              betas=betas,
                              eps=opt_eps,
                              weight_decay=wd)

    scheduler = get_scheduler(sched,
                              optimizer,
                              step=step,
                              gamma=gamma,
                              milestones=milestones,
                              max_decay_steps=mdsteps,
                              end_lr=end_lr,
                              power=power)

    loss_fn, one_hot = get_loss(loss_fn)
    metrics_fn = get_metrics(metrics, nc=nc)

    train_dataloader = torch.utils.data.DataLoader(train_data,
                                                   batch_size=bs,
                                                   shuffle=True)

    valid_dataloader = torch.utils.data.DataLoader(valid_data,
                                                   batch_size=bs,
                                                   shuffle=False)

    test_dataloader = torch.utils.data.DataLoader(test_data,
                                                  batch_size=1,
                                                  shuffle=False)

    if algorithm == 'train':
        train(train_dataloader,
              valid_dataloader,
              model,
              optimizer,
              scheduler,
              loss_fn,
              metrics_fn,
              epochs,
              checkpoints,
              logs,
              device,
              se,
              one_hot,
              nc,
              wandb_flag,
              wandb_metrics)
    elif algorithm == 'test':
        test_logs = test(test_dataloader,
                         model,
                         loss_fn,
                         metrics_fn,
                         logs,
                         device,
                         one_hot,
                         nc,
                         wandb_flag,
                         wandb_metrics)
        if pred:
            indices = random.sample(range(0, len(test_data)), 10)
            os.mkdir(test_logs + '/prediction', 0o777)
            for i in range(10):
                tmp_range = range(nc)
                tmp_dct = {i: names[i] for i in tmp_range}
                prediction(test_data,
                           model,
                           device,
                           tmp_dct,
                           test_logs + f'/prediction/{i}',
                           indices[i])
    else:
        raise ValueError(f'no algorithm {algorithm}')



from .model import Model
from .model import MultiDecModel


def get_decoder(name, encoder, nc, mode, device):
    '''
    Return model object by name of decoder and encoder.

    Parameterts
    -----------
    name : str
        name of decoder.

    encoder : Any
        encoder model.

    nc : int
        classes number.

    mode : str
        mode of decoder ['decoder', 'multidecoder']

    device : str
        name of device ['cpu', 'cuda']
    '''
    if mode != 'multidecoder':
        return Model(name, encoder, nc, device)
    else:
        return MultiDecModel(name, encoder, nc, device)



from torch.utils.data import Dataset
import yaml
import os
from PIL import Image
import numpy as np
import torch


def custom(img_dir, annotation, tr_part, va_part, te_part, transform):
    '''
    Return splited into train, val, test parts your custom dataset.

    Parameterts
    -----------
    img_dir : str
        folder with images.

    annotation : str
        yaml dataset file.

    tr_part : float
        percent of train data.

    va_part : float
        percent of valid data.

    te_part : float
        percent of test data.

    transform : torchvision.transforms
        torchvision transforms object with augmentations.
    '''
    dataset = CustomDataset(img_dir, annotation, transform)

    tr_part = int(tr_part * len(dataset))
    va_part = int(va_part * len(dataset))
    te_part = int(te_part * len(dataset))

    if tr_part + va_part + te_part != len(dataset):
        sizes = [tr_part,
                 va_part,
                 te_part,
                 len(dataset) - tr_part - va_part - te_part]
    else:
        sizes = [tr_part, va_part, te_part]

    datasets = torch.utils.data.random_split(dataset, sizes)
    train_dataset = datasets[0]
    valid_dataset = datasets[1]
    test_dataset = datasets[2]

    return train_dataset, valid_dataset, test_dataset


class CustomDataset(Dataset):
    '''
    Class used to create custom dataset.

    Attributes
    ----------
    images : list[str]
        list with filenames.

    labels : dict[str, int]
        dict for each path to image - class value.

    transforms : Any
        torchvision transforms object with augmentations.

    Methods
    -------
    len()
        calculate length of dataset.

    getitem(idx)
        get dataset element by idx, needed for dataloader.

    load_sample(file)
        load image file into normalized np.array.
    '''

    def __init__(self, img_dir, annotation_file, transform, sobel):
        '''
        Parameters
        ----------
        img_dir : str
            directory with dataset images.

        annotation_file : str
            yaml file with image - class.

        transform : Any
            torchvision transforms object with augmentations.
        '''
        images_files = os.listdir(img_dir)
        self.images = [img_dir + '/' + img for img in images_files]
        with open(annotation_file, 'r') as stream:
            self.labels = yaml.safe_load(stream)
        self.transform = transform

    def load_sample(self, file):
        '''
        Load image file into np.array.

        Parameters
        ----------
        file : str
            image file path.
        '''
        image = Image.open(file)
        return image

    def __len__(self):
        '''
        Len of dataset.
        '''
        return len(self.images)

    def __getitem__(self, idx):
        '''
        Get dataset image, label by idx.

        Parameters
        ----------
        idx : int
            index.
        '''
        label = int(self.labels[self.images[idx]])
        image = self.transform(self.load_sample(self.images[idx])).float()
        return image, label



from .mnist import mnist, fashion_mnist
from .cifar import cifar10, cifar100
from .custom import custom
from .gtsrb import gtsrb


def get_dataset(name,
                train_part,
                valid_part,
                test_part,
                transform,
                datafiles=None):
    '''
    Return splited into train, val, test parts dataset by name.

    Parameterts
    -----------
    name : str
        name of dataset.

    train_part : float
        percent of train data.

    valid_part : float
        percent of valid data.

    test_part : float
        percent of test data.

    transform : torchvision.transforms
        torchvision transforms object with augmentations.
    '''
    if name == 'mnist':
        train_data, valid_data, test_data = mnist(train_part,
                                                  valid_part,
                                                  test_part,
                                                  transform)
    elif name == 'fashion_mnist':
        train_data, valid_data, test_data = fashion_mnist(train_part,
                                                          valid_part,
                                                          test_part,
                                                          transform)
    elif name == 'cifar10':
        train_data, valid_data, test_data = cifar10(train_part,
                                                    valid_part,
                                                    test_part,
                                                    transform)
    elif name == 'cifar100':
        train_data, valid_data, test_data = cifar100(train_part,
                                                     valid_part,
                                                     test_part,
                                                     transform)
    elif name == 'gtsrb':
        train_data, valid_data, test_data = gtsrb(train_part,
                                                  valid_part,
                                                  test_part,
                                                  transform)
    elif name == 'custom':
        train_data, valid_data, test_data = custom(datafiles[0],
                                                   datafiles[1],
                                                   train_part,
                                                   valid_part,
                                                   test_part,
                                                   transform)
    else:
        raise ValueError(f'no dataset {name}')
    return train_data, valid_data, test_data



import torch
from torch.nn import BCELoss, BCEWithLogitsLoss
from torch import nn


def bce():
    '''
    Return binary-cross-entropy loss object.
    https://pytorch.org/docs/stable/generated/torch.nn.BCELoss.html
    '''
    return BCELoss()


def bcelogits():
    '''
    Return binary-cross-entropy-with-logits loss object.
    https://pytorch.org/docs/stable/generated/torch.nn.BCEWithLogitsLoss.html
    '''
    return BCEWithLogitsLoss()


def custombce():
    '''
    Return custom binary-cross-entropy loss object from CustomBCELoss class.
    https://pytorch.org/docs/stable/generated/torch.nn.BCELoss.html
    '''
    return CustomBCELoss()


def custombcelogits():
    '''
    Return custom binary-cross-entropy-with-logits loss object
    from CustomBCEWithLogitsLoss class.
    https://pytorch.org/docs/stable/generated/torch.nn.BCEWithLogitsLoss.html
    '''
    return CustomBCEWithLogitsLoss()


class CustomBCELoss(nn.Module):
    '''
    Class used to calculate binary-cross-entropy loss.

    Attributes
    ----------
    weight : float, list[float]
        weight for loss value.

    reduction : str
        type of reduction ['mean', 'sum'].

    Methods
    -------
    forward(inputs, targets)
        calculate binary-cross-entropy loss between inputs and targets.
    '''

    def __init__(self, weight=1.0, reduction='mean'):
        '''
        Parameters
        ----------
        weight : float, list[float], optional
            weight for loss value.

        reduction : str, optional
            type of reduction ['mean', 'sum'].
        '''
        super().__init__()
        self.weight = weight
        self.reduction = reduction

    def forward(self, inputs, targets):
        '''
        Calculate binary-cross-entropy loss between inputs and targets.

        Parameterts
        -----------
        inputs : torch.tensor
            model predictions.

        targets : torch.tensor
            ground-truth labels.
        '''
        inputs = inputs.view(-1)
        targets = targets.view(-1)
        logx = torch.clamp(torch.log2(inputs), min=-100.0, max=100.0)
        loginvx = torch.clamp(torch.log2(1.0 - inputs), min=-100.0, max=100.0)
        loss = -self.weight * (targets * logx + (1.0 - targets) * loginvx)
        if self.reduction == 'mean':
            return loss.mean()
        elif self.reduction == 'sum':
            return loss.sum()


class CustomBCEWithLogitsLoss(nn.Module):
    '''
    Class used to calculate binary-cross-entropy-with-logits loss.

    Attributes
    ----------
    weight : float, list[float]
        weight for loss value.

    reduction : str
        type of reduction ['mean', 'sum'].

    sigm : Any
        sigmoid activation function.

    Methods
    -------
    forward(inputs, targets)
        calculate binary-cross-entropy-with-logits loss
        between inputs and targets.
    '''

    def __init__(self, weight=1.0, reduction='mean'):
        '''
        Parameters
        ----------
        weight : float, list[float], optional
            weight for loss value.

        reduction : str, optional
            type of reduction ['mean', 'sum'].
        '''
        super().__init__()
        self.weight = weight
        self.reduction = reduction
        self.sigm = nn.Sigmoid()

    def forward(self, inputs, targets):
        '''
        Calculate binary-cross-entropy-with-logits loss
        between inputs and targets.

        Parameterts
        -----------
        inputs : torch.tensor
            model predictions.

        targets : torch.tensor
            ground-truth labels.
        '''
        inputs = inputs.view(-1)
        targets = targets.view(-1)
        sigm_inps = self.sigm(inputs)
        logx = torch.clamp(torch.log2(sigm_inps), min=-100.0, max=100.0)
        loginvx = torch.clamp(torch.log2(1.0 - sigm_inps),
                              min=-100.0, max=100.0)
        loss = -self.weight * (targets * logx + (1.0 - targets) * loginvx)
        if self.reduction == 'mean':
            return loss.mean()
        elif self.reduction == 'sum':
            return loss.sum()



from .celoss import celoss, customceloss
from .bceloss import bce, bcelogits, custombce, custombcelogits
from .meanloss import mse, mae, custom_mse, custom_mae


def get_loss(loss_fn):
    '''
    Return loss_fn object by name and one_hot flag.

    Parameterts
    -----------
    loss_fn : str
        name of loss function.
    '''
    one_hot = False
    if loss_fn == 'ce':
        loss_fn = celoss()
    elif loss_fn == 'custom_ce':
        loss_fn = customceloss()
    elif loss_fn == 'bce':
        loss_fn = bce()
        one_hot = True
    elif loss_fn == 'custom_bce':
        loss_fn = custombce()
        one_hot = True
    elif loss_fn == 'bce_logits':
        loss_fn = bcelogits()
        one_hot = True
    elif loss_fn == 'custom_bce_logits':
        loss_fn = custombcelogits()
        one_hot = True
    elif loss_fn == 'mse':
        loss_fn = mse()
        one_hot = True
    elif loss_fn == 'custom_mse':
        loss_fn = custom_mse()
        one_hot = True
    elif loss_fn == 'mae':
        loss_fn = mae()
        one_hot = True
    elif loss_fn == 'custom_mae':
        loss_fn = custom_mae()
        one_hot = True
    else:
        raise ValueError(f'no loss function {loss_fn}')
    return loss_fn, one_hot



from .alexnet import AlexnetFeatures
from .lenet import LenetFeatures
from .resnet import Resnet18Features, \
    Resnet34Features, \
    Resnet50Features, \
    Resnet101Features, \
    Resnet152Features
from .googlenet import GoognetNoAuxFeatures
from .mobilenet import MobilenetFeatures
from .vgg import VGG11Features, VGG16Features, VGG19Features
from .base import get_encoder



import torch


def prediction(dataset, model, device, class_name, save_path, ind):
    '''
    Predict result for one image.

    Parameterts
    -----------
    dataset : list[torch.tensor, int]
        dataset.

    device : str
        device ['cpu', 'cuda'].

    class_name : list[str]
        name for all classes.

    save_path : str
        path to save output file

    ind : int
        index of image in dataset
    '''
    model.eval()
    img = dataset[ind][0]
    label = dataset[ind][1]
    img = img.unsqueeze(0).to(device)
    pred = torch.argmax(model(img), dim=1).cpu()[0]

    with open(save_path + '.log', 'w+') as in_f:
        in_f.write(f'predict class {class_name[pred.item()]}\n')
        in_f.write(f'correct class {class_name[label]}\n')



from .alexnet import AlexnetFeatures
from .lenet import LenetFeatures
from .resnet import Resnet18Features, \
    Resnet34Features, \
    Resnet50Features, \
    Resnet101Features, \
    Resnet152Features
from .googlenet import GoognetNoAuxFeatures
from .mobilenet import MobilenetFeatures
from .vgg import VGG11Features, VGG16Features, VGG19Features


def get_encoder(name, inc):
    '''
    Return encoder object by name.

    Parameterts
    -----------
    name : str
        name of encoder.

    inc : int
        number of input channels [1, 3, 4].
    '''
    if name == 'alexnet':
        return AlexnetFeatures(inc)
    elif name == 'lenet':
        return LenetFeatures(inc)
    elif name == 'resnet18':
        return Resnet18Features(inc)
    elif name == 'resnet34':
        return Resnet34Features(inc)
    elif name == 'resnet50':
        return Resnet50Features(inc)
    elif name == 'resnet101':
        return Resnet101Features(inc)
    elif name == 'resnet152':
        return Resnet152Features(inc)
    elif name == 'googlenet':
        return GoognetNoAuxFeatures(inc)
    elif name == 'mobilenet':
        return MobilenetFeatures(inc)
    elif name == 'vgg11':
        return VGG11Features(inc)
    elif name == 'vgg16':
        return VGG16Features(inc)
    elif name == 'vgg19':
        return VGG19Features(inc)
    else:
        raise ValueError(f'no encoder {name}')



#!/usr/bin/env python
# pip install twine

import io
import os
import sys
from shutil import rmtree
from setuptools import setup, Command, find_packages

NAME = 'opennn_pytorch'
DESCRIPTION = 'Neural Networks library for image classification task.'
URL = 'https://github.com/Pe4enIks/OpenNN'
EMAIL = 'evgeniipishchik@mail.ru'
AUTHOR = 'Evgenii Pishchik'
REQUIRES_PYTHON = '>=3.9.0'
VERSION = None

here = os.path.abspath(os.path.dirname(__file__))

with open(os.path.join(here, 'requirments.txt'), encoding='utf-8') as f:
    REQUIRED = f.read().split('\n')
print(REQUIRED)

try:
    with io.open(os.path.join(here, 'README.md'), encoding='utf-8') as f:
        long_description = '\n' + f.read()
except FileNotFoundError:
    long_description = DESCRIPTION

about = {}
if not VERSION:
    with open(os.path.join(here, NAME, '__version__.py')) as f:
        exec(f.read(), about)
else:
    about['__version__'] = VERSION
DOWNLOAD_URL = f'https://github.com/Pe4enIks/OpenNN/archive/refs/tags/v{about["__version__"]}.tar.gz'


class UploadCommand(Command):
    '''Support setup.py upload.'''

    description = 'Build and publish the package.'
    user_options = []

    @staticmethod
    def status(s):
        '''Prints things in bold.'''
        print('\033[1m{0}\033[0m'.format(s))

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        try:
            self.status('Removing previous builds…')
            rmtree(os.path.join(here, 'dist'))
        except OSError:
            pass

        self.status('Building Source and Wheel (universal) distribution…')
        os.system('{0} setup.py sdist bdist_wheel --universal'.format(sys.executable))

        self.status('Uploading the package to PyPI via Twine…')
        os.system('twine upload dist/*')

        self.status('Pushing git tags…')
        os.system('git tag v{0}'.format(about['__version__']))
        os.system('git push --tags')

        sys.exit()


setup(
    name=NAME,
    version=about['__version__'],
    description=DESCRIPTION,
    long_description=long_description,
    long_description_content_type='text/markdown',
    author=AUTHOR,
    author_email=EMAIL,
    python_requires=REQUIRES_PYTHON,
    url=URL,
    download_url=DOWNLOAD_URL,
    license='MIT',
    classifiers=[
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.9',
    ],
    py_modules=[f'{NAME}'],
    packages=find_packages(),
    install_requires=REQUIRED,
    include_package_data=True,
    cmdclass={
        'upload': UploadCommand,
    },
)



from . import algo
from . import losses
from . import metrics
from . import datasets
from . import optimizers
from . import encoders
from . import decoders
from .run import run, transforms_lst, parse_yaml



from torch.optim.lr_scheduler import StepLR, MultiStepLR


def steplr(optimizer, step, gamma):
    '''
    Return StepLR scheduler object.
    https://pytorch.org/docs/stable/generated/torch.optim.lr_scheduler.StepLR.html

    Parameterts
    -----------
    optimizer : Optimizer
        optimizer object.

    step : int
        every step epochs do scheduler step.

    gamma : float
        lr will be multiplied to this value each step epochs.
    '''
    scheduler = StepLR(optimizer, step, gamma)
    return scheduler


def multisteplr(optimizer, milestones, gamma):
    '''
    Return MultiStepLR scheduler object.
    https://pytorch.org/docs/stable/generated/torch.optim.lr_scheduler.MultiStepLR.html

    Parameterts
    -----------
    optimizer : Optimizer
        optimizer object.

    milestones : list[int]
        checkpoints for lr * gamma.

    gamma : float
        lr will be multiplied to this value each step epochs.
    '''
    scheduler = MultiStepLR(optimizer, milestones, gamma)
    return scheduler



from torch.nn import MSELoss, L1Loss
from torch import nn
import torch


def mse():
    '''
    Return mean-squared-error loss object.
    https://pytorch.org/docs/stable/generated/torch.nn.MSELoss.html
    '''
    return MSELoss()


def custom_mse():
    '''
    Return mean-squared-error loss object from CustomMSELoss class.
    https://pytorch.org/docs/stable/generated/torch.nn.MSELoss.html
    '''
    return CustomMSELoss()


def mae():
    '''
    Return mean-absolute-error loss object.
    https://pytorch.org/docs/stable/generated/torch.nn.L1Loss.html
    '''
    return L1Loss()


def custom_mae():
    '''
    Return mean-absolute-error loss object from CustomMAELoss class.
    https://pytorch.org/docs/stable/generated/torch.nn.L1Loss.html
    '''
    return CustomMAELoss()


class CustomMSELoss(nn.Module):
    '''
    Class used to calculate mean-squared-error loss.

    Attributes
    ----------
    reduction : str
        type of reduction ['mean', 'sum'].

    Methods
    -------
    forward(inputs, targets)
        calculate mean-squared-error loss between inputs and targets.
    '''
    def __init__(self, reduction='mean'):
        '''
        Parameters
        ----------
        reduction : str, optional
            type of reduction ['mean', 'sum'].
        '''
        super().__init__()
        self.reduction = reduction

    def forward(self, inputs, targets):
        '''
        Calculate mean-squared-error loss between inputs and targets.

        Parameterts
        -----------
        inputs : torch.tensor
            model predictions.

        targets : torch.tensor
            ground-truth labels.
        '''
        inputs = inputs.view(-1)
        targets = targets.view(-1)
        sub = (inputs - targets) ** 2
        if self.reduction == 'mean':
            return sub.mean()
        elif self.reduction == 'sum':
            return sub.sum()


class CustomMAELoss(nn.Module):
    '''
    Class used to calculate mean-absolute-error loss.

    Attributes
    ----------
    reduction : str
        type of reduction ['mean', 'sum'].

    Methods
    -------
    forward(inputs, targets)
        calculate mean-absolute-error loss between inputs and targets.
    '''
    def __init__(self, reduction='mean'):
        '''
        Parameters
        ----------
        reduction : str, optional
            type of reduction ['mean', 'sum'].
        '''
        super().__init__()
        self.reduction = reduction

    def forward(self, inputs, targets):
        '''
        Calculate mean-absolute-error loss between inputs and targets.

        Parameterts
        -----------
        inputs : torch.tensor
            model predictions.

        targets : torch.tensor
            ground-truth labels.
        '''
        inputs = inputs.view(-1)
        targets = targets.view(-1)
        sub = torch.abs(inputs - targets)
        if self.reduction == 'mean':
            return sub.mean()
        elif self.reduction == 'sum':
            return sub.sum()



from torch import nn


class LenetFeatures(nn.Module):
    '''
    Class used to calculate features using lenet architecture.
    http://vision.stanford.edu/cs598_spring07/papers/Lecun98.pdf

    Methods
    -------
    forward(x)
        calculate features from input image x.

    out_features()
        return number of output features, needed for decoders input channels.

    name()
        return name of encoder.
    '''

    def __init__(self, inc):
        '''
        Create lenet encoder layers.

        Parameters
        ----------
        inc : int
            number of input channels [1, 3, 4].
        '''
        super().__init__()
        self.features = 400
        self.conv1 = nn.Conv2d(inc, 6, kernel_size=(5, 5), padding='same')
        self.conv2 = nn.Conv2d(6, 16, kernel_size=(5, 5))
        self.maxpool = nn.MaxPool2d(kernel_size=2,
                                    stride=2,
                                    padding=0,
                                    dilation=1,
                                    ceil_mode=False)
        self.avgpool = nn.AdaptiveAvgPool2d(output_size=(5, 5))
        self.tanh = nn.Tanh()

    def forward(self, x):
        '''
        Calculate features from input image x.

        Parameterts
        -----------
        x : torch.tensor
            input image.
        '''
        b1 = self.maxpool(self.tanh(self.conv1(x)))
        b2 = self.avgpool(self.maxpool(self.tanh(self.conv2(b1))))
        return b2

    def out_features(self):
        '''
        Return number of output features, needed for decoders input channels.
        '''
        return self.features

    def name(self):
        '''
        Return name of encoder.
        '''
        return 'lenet'
